# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Forecast output."""

from functools import partial

import pandas as pd
from sklearn.metrics import mean_squared_error

from greykite.common import constants as cst
from greykite.common.evaluation import ElementwiseEvaluationMetricEnum
from greykite.common.evaluation import EvaluationMetricEnum
from greykite.common.evaluation import add_finite_filter_to_scorer
from greykite.common.evaluation import calc_pred_coverage
from greykite.common.evaluation import calc_pred_err
from greykite.common.evaluation import fraction_outside_tolerance
from greykite.common.evaluation import r2_null_model_score
from greykite.common.python_utils import apply_func_to_columns
from greykite.common.viz.timeseries_plotting import add_groupby_column
from greykite.common.viz.timeseries_plotting import flexible_grouping_evaluation
from greykite.common.viz.timeseries_plotting import grouping_evaluation
from greykite.common.viz.timeseries_plotting import plot_forecast_vs_actual
from greykite.common.viz.timeseries_plotting import plot_multivariate
from greykite.common.viz.timeseries_plotting import plot_univariate
from greykite.framework.input.univariate_time_series import UnivariateTimeSeries


class UnivariateForecast:
    """Stores predicted and actual values.
    Provides functionality to evaluate a forecast:

        - plots true against actual with prediction bands.
        - evaluates model performance.

    Input should be one of two kinds of forecast results:

        - model fit to train data, forecast on test set (actuals available).
        - model fit to all data, forecast on future dates (actuals not available).

    The input ``df`` is a concatenation of fitted and forecasted values.

    Attributes
    ----------
    df : `pandas.DataFrame`
        Timestamp, predicted, and actual values.
    time_col : `str`
        Column in ``df`` with timestamp (default "ts").
    actual_col : `str`
        Column in ``df`` with actual values (default "y").
    predicted_col : `str`
        Column in ``df`` with predicted values (default "forecast").
    predicted_lower_col : `str` or None
        Column in ``df`` with predicted lower bound (default "forecast_lower", optional).
    predicted_upper_col : `str` or None
        Column in ``df`` with predicted upper bound (default "forecast_upper", optional).
    null_model_predicted_col : `str` or None
        Column in ``df`` with predicted value of null model (default "forecast_null", optional).
    ylabel : `str`
        Unit of measurement (default "y")
    train_end_date : `str` or `datetime` or None, default None
        End date for train period. If `None`, assumes all data were used for training.
    test_start_date : `str` or `datetime` or None, default None
        Start date of test period. If `None`, set to the ``time_col`` value immediately after
        ``train_end_date``. This assumes that all data not used in training were used for testing.
    forecast_horizon : `int` or None, default None
        Number of periods forecasted into the future. Must be > 0.
    coverage : `float` or None
        Intended coverage of the prediction bands (0.0 to 1.0).
    r2_loss_function : `function`
        Loss function to calculate ``cst.R2_null_model_score``, with signature
        ``loss_func(y_true, y_pred)`` (default mean_squared_error)
    estimator : An instance of an estimator that implements `greykite.models.base_forecast_estimator.BaseForecastEstimator`.
        The fitted estimator, the last step in the forecast pipeline.
    relative_error_tolerance : `float` or None, default None
        Threshold to compute the ``Outside Tolerance`` metric,
        defined as the fraction of forecasted values whose relative
        error is strictly greater than ``relative_error_tolerance``.
        For example, 0.05 allows for 5% relative error.
        If `None`, the metric is not computed.
    df_train : `pandas.DataFrame`
        Subset of ``df`` where ``df[time_col]`` <= ``train_end_date``.
    df_test : `pandas.DataFrame`
        Subset of ``df`` where ``df[time_col]`` > ``train_end_date``.
    train_evaluation : `dict` [`str`, `float`]
        Evaluation metrics on training set.
    test_evaluation : `dict` [`str`, `float`]
        Evaluation metrics on test set (if actual values provided after train_end_date).
    test_na_count : `int`
        Count of NA values in test data.
    """

    def __init__(
            self,
            df,
            time_col=cst.TIME_COL,
            actual_col=cst.ACTUAL_COL,
            predicted_col=cst.PREDICTED_COL,
            predicted_lower_col=cst.PREDICTED_LOWER_COL,
            predicted_upper_col=cst.PREDICTED_UPPER_COL,
            null_model_predicted_col=cst.NULL_PREDICTED_COL,
            ylabel=cst.VALUE_COL,
            train_end_date=None,
            test_start_date=None,
            forecast_horizon=None,
            coverage=0.95,
            r2_loss_function=mean_squared_error,
            estimator=None,
            relative_error_tolerance=None):
        if predicted_lower_col is not None or predicted_upper_col is not None:
            if coverage is None:
                raise ValueError("`coverage` must be provided when lower/upper bounds are set")
            elif coverage < 0.0 or coverage > 1.0:
                raise ValueError("`coverage` must be between 0.0 and 1.0")

        if train_end_date is not None and not all(pd.Series(train_end_date).isin(df[time_col])):
            raise ValueError(
                f"train_end_date {train_end_date} is not found in time column.\n"
                f"The time range in data is: from {min(df[time_col])} to {max(df[time_col])}.\n"
                f"The data size is: {df.shape[0]}.\n"
                f"Type of train_end_date is {type(train_end_date)}.\n"
                f"Type of df[time_col] is {df[time_col].dtype}.\n")

        if any([col not in df.columns
                for col in [time_col, actual_col, predicted_col, predicted_lower_col, predicted_upper_col,
                            null_model_predicted_col] if col is not None]):
            raise ValueError(f"Column not found in data frame")

        self.df = df
        self.time_col = time_col
        self.actual_col = actual_col
        self.predicted_col = predicted_col
        self.predicted_lower_col = predicted_lower_col
        self.predicted_upper_col = predicted_upper_col
        self.null_model_predicted_col = null_model_predicted_col
        self.ylabel = ylabel
        self.train_end_date = train_end_date
        self.test_start_date = test_start_date
        self.forecast_horizon = forecast_horizon
        self.coverage = coverage
        self.r2_loss_function = r2_loss_function
        self.estimator = estimator
        self.relative_error_tolerance = relative_error_tolerance

        self.df[self.time_col] = pd.to_datetime(df[time_col], infer_datetime_format=True)
        if train_end_date is None:
            self.train_end_date = df[self.time_col].max()
        else:
            self.train_end_date = pd.to_datetime(train_end_date, infer_datetime_format=True)
        self.df_train = df[df[time_col] <= self.train_end_date]

        if test_start_date is None:
            # This expects no gaps in time column
            inferred_freq = pd.infer_freq(self.df[self.time_col])
            # Uses pd.date_range because pd.Timedelta does not work for complicated frequencies e.g. "W-MON"
            self.test_start_date = pd.date_range(
                start=self.train_end_date,
                periods=2,
                freq=inferred_freq)[-1]
        else:
            self.test_start_date = pd.to_datetime(test_start_date, infer_datetime_format=True)
        self.df_test = df[df[time_col] >= self.test_start_date]

        self.test_na_count = self.df_test[actual_col].isna().sum()

        # compute evaluation metrics
        self.train_evaluation = None
        self.test_evaluation = None
        self.compute_evaluation_metrics_split()

    def __evaluation_metrics(self, data):
        """Computes various evaluation metrics for the forecast.

        :param data: pd.DataFrame
            with columns according to self.__init__ (e.g. subset of self.df)
        :return: dictionary with evaluation metrics
        """
        metrics = calc_pred_err(data[self.actual_col], data[self.predicted_col])
        if self.relative_error_tolerance is not None:
            metrics[cst.FRACTION_OUTSIDE_TOLERANCE] = partial(
                fraction_outside_tolerance,
                rtol=self.relative_error_tolerance)(
                    y_true=data[self.actual_col],
                    y_pred=data[self.predicted_col])
        else:
            metrics[cst.FRACTION_OUTSIDE_TOLERANCE] = None

        for metric in [cst.R2_null_model_score, cst.PREDICTION_BAND_WIDTH, cst.PREDICTION_BAND_COVERAGE,
                       cst.LOWER_BAND_COVERAGE, cst.UPPER_BAND_COVERAGE, cst.COVERAGE_VS_INTENDED_DIFF]:
            metrics.update({metric: None})

        # evaluates cst.R2_null_model_score if null model is available
        if self.null_model_predicted_col is not None:
            metrics.update({cst.R2_null_model_score: r2_null_model_score(data[self.actual_col],
                                                                         data[self.predicted_col],
                                                                         y_pred_null=data[self.null_model_predicted_col],
                                                                         loss_func=self.r2_loss_function)})

        # evaluates prediction bands if available
        if self.predicted_lower_col is not None and self.predicted_upper_col is not None:
            result = calc_pred_coverage(
                data[self.actual_col],
                data[self.predicted_col],
                data[self.predicted_lower_col],
                data[self.predicted_upper_col],
                self.coverage)
            metrics.update(result)
        return metrics

    def compute_evaluation_metrics_split(self):
        """Computes __evaluation_metrics for train and test set separately.

        :return: dictionary with train and test evaluation metrics
        """
        self.train_evaluation = self.__evaluation_metrics(self.df_train)
        # only test evaluation if there are actuals (e.g. not future prediction and not all missing)
        if self.df_test.shape[0] > 0 and self.test_na_count < self.df_test.shape[0]:
            self.test_evaluation = self.__evaluation_metrics(self.df_test)
        return {
            "Train": self.train_evaluation,
            "Test": self.test_evaluation
        }

    def plot(self, **kwargs):
        """Plots predicted against actual.

        Parameters
        ----------
        kwargs : additional parameters
            Additional parameters to pass to
            `~greykite.common.viz.timeseries_plotting.plot_forecast_vs_actual`
            such as title, colors, and line styling.

        Returns
        -------
        fig : `plotly.graph_objs.Figure`
            Plotly figure of forecast against actuals, with prediction
            intervals if available.

            See `~greykite.common.viz.timeseries_plotting.plot_forecast_vs_actual`
            return value for how to plot the figure and add customization.
        """
        return plot_forecast_vs_actual(
            self.df,
            time_col=self.time_col,
            actual_col=self.actual_col,
            predicted_col=self.predicted_col,
            predicted_lower_col=self.predicted_lower_col,
            predicted_upper_col=self.predicted_upper_col,
            ylabel=self.ylabel,
            train_end_date=self.train_end_date,
            **kwargs)

    def get_grouping_evaluation(
            self,
            score_func=EvaluationMetricEnum.MeanAbsolutePercentError.get_metric_func(),
            score_func_name=EvaluationMetricEnum.MeanAbsolutePercentError.get_metric_name(),
            which="train",
            groupby_time_feature=None,
            groupby_sliding_window_size=None,
            groupby_custom_column=None):
        """Group-wise computation of forecasting error.
        Can be used to evaluate error/ aggregated value by a time feature,
        over time, or by a user-provided column.

        Exactly one of: ``groupby_time_feature``, ``groupby_sliding_window_size``,
        ``groupby_custom_column`` must be provided.

        Parameters
        ----------
        score_func : callable, optional
            Function that maps two arrays to a number.
            Signature (y_true: array, y_pred: array) -> error: float
        score_func_name : `str` or None, optional
            Name of the score function used to report results.
            If None, defaults to "metric".
        which: `str`
            "train" or "test". Which dataset to evaluate.
        groupby_time_feature : `str` or None, optional
            If provided, groups by a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        groupby_sliding_window_size : `int` or None, optional
            If provided, sequentially partitions data into groups of size
            ``groupby_sliding_window_size``.
        groupby_custom_column : `pandas.Series` or None, optional
            If provided, groups by this column value. Should be same length as the DataFrame.

        Returns
        -------
        grouped_df : `pandas.DataFrame` with two columns:

            (1) grouping_func_name:
                evaluation metric computing forecasting error of timeseries.
            (2) group name:
                group name depends on the grouping method:
                ``groupby_time_feature`` for ``groupby_time_feature``
                ``cst.TIME_COL`` for ``groupby_sliding_window_size``
                ``groupby_custom_column.name`` for ``groupby_custom_column``.
        """
        df = self.df_train.copy() if which.lower() == "train" else self.df_test.copy()
        score_func = add_finite_filter_to_scorer(score_func)  # in case it's not already added
        if score_func_name:
            grouping_func_name = f"{which} {score_func_name}"
        else:
            grouping_func_name = f"{which} metric"

        def grouping_func(grp):
            return score_func(grp[self.actual_col], grp[self.predicted_col])

        result = add_groupby_column(
            df=df,
            time_col=self.time_col,
            groupby_time_feature=groupby_time_feature,
            groupby_sliding_window_size=groupby_sliding_window_size,
            groupby_custom_column=groupby_custom_column)

        grouped_df = grouping_evaluation(
            df=result["df"],
            groupby_col=result["groupby_col"],
            grouping_func=grouping_func,
            grouping_func_name=grouping_func_name)
        return grouped_df

    def plot_grouping_evaluation(
            self,
            score_func=EvaluationMetricEnum.MeanAbsolutePercentError.get_metric_func(),
            score_func_name=EvaluationMetricEnum.MeanAbsolutePercentError.get_metric_name(),
            which="train",
            groupby_time_feature=None,
            groupby_sliding_window_size=None,
            groupby_custom_column=None,
            xlabel=None,
            ylabel=None,
            title=None):
        """Computes error by group and plots the result.
        Can be used to plot error by a time feature, over time, or by a user-provided column.

        Exactly one of: ``groupby_time_feature``, ``groupby_sliding_window_size``,
        ``groupby_custom_column`` must be provided.

        Parameters
        ----------
        score_func : callable, optional
            Function that maps two arrays to a number.
            Signature (y_true: array, y_pred: array) -> error: float
        score_func_name : `str` or None, optional
            Name of the score function used to report results.
            If None, defaults to "metric".
        which: `str`, optional, default "train"
            Which dataset to evaluate, "train" or "test".
        groupby_time_feature : `str` or None, optional
            If provided, groups by a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        groupby_sliding_window_size : `int` or None, optional
            If provided, sequentially partitions data into groups of size
            ``groupby_sliding_window_size``.
        groupby_custom_column : `pandas.Series` or None, optional
            If provided, groups by this column value. Should be same length as the DataFrame.
        xlabel : `str`, optional, default None
            X-axis label of the plot.
        ylabel : `str`, optional, default None
            Y-axis label of the plot.
        title : `str` or None, optional
            Plot title, if None this function creates a suitable title.

        Returns
        -------
        fig : `plotly.graph_objs.Figure`
            plotly graph object showing forecasting error by group.
            x-axis label depends on the grouping method:
            ``groupby_time_feature`` for ``groupby_time_feature``
            ``time_col`` for ``groupby_sliding_window_size``
            ``groupby_custom_column.name`` for ``groupby_custom_column``.
        """
        grouped_df = self.get_grouping_evaluation(
            score_func=score_func,
            score_func_name=score_func_name,
            which=which,
            groupby_time_feature=groupby_time_feature,
            groupby_sliding_window_size=groupby_sliding_window_size,
            groupby_custom_column=groupby_custom_column)

        xcol, ycol = grouped_df.columns
        fig = plot_univariate(
            df=grouped_df,
            x_col=xcol,
            y_col=ycol,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title)

        return fig

    def autocomplete_map_func_dict(self, map_func_dict):
        """Sweeps through ``map_func_dict``, converting values that are
        `~greykite.common.evaluation.ElementwiseEvaluationMetricEnum`
        member names to their corresponding row-wise evaluation function with appropriate
        column names for this UnivariateForecast instance.

        For example::

            map_func_dict = {
                "squared_error": ElementwiseEvaluationMetricEnum.SquaredError.name,
                "coverage": ElementwiseEvaluationMetricEnum.Coverage.name,
                "custom_metric": custom_function
            }

            is converted to

            map_func_dict = {
                "squared_error": lambda row: ElementwiseEvaluationMetricEnum.SquaredError.get_metric_func()(
                                            row[self.actual_col],
                                            row[self.predicted_col]),
                "coverage": lambda row: ElementwiseEvaluationMetricEnum.Coverage.get_metric_func()(
                                            row[self.actual_col],
                                            row[self.predicted_lower_col],
                                            row[self.predicted_upper_col]),
                "custom_metric": custom_function
            }

        Parameters
        ----------
        map_func_dict : `dict` or None
            Same as `~greykite.common.viz.timeseries_plotting.flexible_grouping_evaluation`,
            with one exception: values may a ElementwiseEvaluationMetricEnum member name.
            There are converted a callable for ``flexible_grouping_evaluation``.

        Returns
        -------
        map_func_dict : `dict`
            Can be passed to `~greykite.common.viz.timeseries_plotting.flexible_grouping_evaluation`.
        """
        if map_func_dict is None:
            updated_map_func_dict = None
        else:
            updated_map_func_dict = {}
            for name, func in map_func_dict.items():
                if isinstance(func, str):
                    if func in ElementwiseEvaluationMetricEnum.__members__:
                        enum = ElementwiseEvaluationMetricEnum[func]
                        row_func = enum.get_metric_func()
                        args = enum.get_metric_args()
                        arg_mapping = {  # maps canonical names to the names in self.df
                            cst.ACTUAL_COL: self.actual_col,
                            cst.PREDICTED_COL: self.predicted_col,
                            cst.PREDICTED_LOWER_COL: self.predicted_lower_col,
                            cst.PREDICTED_UPPER_COL: self.predicted_upper_col,
                            cst.NULL_PREDICTED_COL: self.null_model_predicted_col,
                            cst.TIME_COL: self.time_col,
                        }
                        try:
                            cols = [arg_mapping[arg] for arg in args]
                        except KeyError:
                            raise ValueError(f"Auto-complete for {func} is not "
                                             f"enabled, please specify the full function.")

                        # Creates a function that passes the appropriate columns
                        # to the elementwise evaluation function
                        updated_map_func_dict[name] = apply_func_to_columns(row_func, cols)
                    else:
                        valid_names = ", ".join(ElementwiseEvaluationMetricEnum.__dict__["_member_names_"])
                        raise ValueError(f"{func} is not a recognized elementwise evaluation "
                                         f"metric. Must be one of: {valid_names}")
                else:
                    # `func` is a callable
                    updated_map_func_dict[name] = func
        return updated_map_func_dict

    def get_flexible_grouping_evaluation(
            self,
            which="train",
            groupby_time_feature=None,
            groupby_sliding_window_size=None,
            groupby_custom_column=None,
            map_func_dict=None,
            agg_kwargs=None,
            extend_col_names=False):
        """Group-wise computation of evaluation metrics. Whereas ``self.get_grouping_evaluation``
        computes one metric, this allows computation of any number of custom metrics.

        For example:

            * Mean and quantiles of squared error by group.
            * Mean and quantiles of residuals by group.
            * Mean and quantiles of actual and forecast by group.
            * % of actuals outside prediction intervals by group
            * any combination of the above metrics by the same group

        First adds a groupby column by passing ``groupby_`` parameters to
        `~greykite.common.viz.timeseries_plotting.add_groupby_column`.
        Then computes grouped evaluation metrics by passing ``map_func_dict``,
        ``agg_kwargs`` and ``extend_col_names`` to
        `~greykite.common.viz.timeseries_plotting.flexible_grouping_evaluation`.

        Exactly one of: ``groupby_time_feature``, ``groupby_sliding_window_size``,
        ``groupby_custom_column`` must be provided.

        which: `str`
            "train" or "test". Which dataset to evaluate.
        groupby_time_feature : `str` or None, optional
            If provided, groups by a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        groupby_sliding_window_size : `int` or None, optional
            If provided, sequentially partitions data into groups of size
            ``groupby_sliding_window_size``.
        groupby_custom_column : `pandas.Series` or None, optional
            If provided, groups by this column value. Should be same length as the DataFrame.
        map_func_dict : `dict` [`str`, `callable`] or None, default None
            Row-wise transformation functions to create new columns.
            If None, no new columns are added.

                - key: new column name
                - value: row-wise function to apply to ``df`` to generate the column value.
                         Signature (row: `pandas.DataFrame`) -> transformed value: `float`.

            For example::

                map_func_dict = {
                    "residual": lambda row: row["actual"] - row["forecast"],
                    "squared_error": lambda row: (row["actual"] - row["forecast"])**2
                }

            Some predefined functions are available in
            `~greykite.common.evaluation.ElementwiseEvaluationMetricEnum`. For example::

                map_func_dict = {
                    "residual": lambda row: ElementwiseEvaluationMetricEnum.Residual.get_metric_func()(
                        row["actual"],
                        row["forecast"]),
                    "squared_error": lambda row: ElementwiseEvaluationMetricEnum.SquaredError.get_metric_func()(
                        row["actual"],
                        row["forecast"]),
                    "q90_loss": lambda row: ElementwiseEvaluationMetricEnum.Quantile90.get_metric_func()(
                        row["actual"],
                        row["forecast"]),
                    "abs_percent_error": lambda row: ElementwiseEvaluationMetricEnum.AbsolutePercentError.get_metric_func()(
                        row["actual"],
                        row["forecast"]),
                    "coverage": lambda row: ElementwiseEvaluationMetricEnum.Coverage.get_metric_func()(
                        row["actual"],
                        row["forecast_lower"],
                        row["forecast_upper"]),
                }

            As shorthand, it is sufficient to provide the enum member name.  These are
            auto-expanded into the appropriate function.
            So the following is equivalent::

                map_func_dict = {
                    "residual": ElementwiseEvaluationMetricEnum.Residual.name,
                    "squared_error": ElementwiseEvaluationMetricEnum.SquaredError.name,
                    "q90_loss": ElementwiseEvaluationMetricEnum.Quantile90.name,
                    "abs_percent_error": ElementwiseEvaluationMetricEnum.AbsolutePercentError.name,
                    "coverage": ElementwiseEvaluationMetricEnum.Coverage.name,
                }

        agg_kwargs : `dict` or None, default None
            Passed as keyword args to `pandas.core.groupby.DataFrameGroupBy.aggregate` after creating
            new columns and grouping by ``groupby_col``.

            See `pandas.core.groupby.DataFrameGroupBy.aggregate` or
            `~greykite.common.viz.timeseries_plotting.flexible_grouping_evaluation`
            for details.

        extend_col_names : `bool` or None, default False
            How to flatten index after aggregation.
            In some cases, the column index after aggregation is a multi-index.
            This parameter controls how to flatten an index with 2 levels to 1 level.

                - If None, the index is not flattened.
                - If True, column name is a composite: ``{index0}_{index1}``
                  Use this option if index1 is not unique.
                - If False, column name is simply ``{index1}``

            Ignored if the ColumnIndex after aggregation has only one level (e.g.
            if named aggregation is used in ``agg_kwargs``).

        Returns
        -------
        df_transformed : `pandas.DataFrame`
            ``df`` after transformation and optional aggregation.

            If ``groupby_col`` is None, returns ``df`` with additional columns as the keys in ``map_func_dict``.
            Otherwise, ``df`` is grouped by ``groupby_col`` and this becomes the index. Columns
            are determined by ``agg_kwargs`` and ``extend_col_names``.

        See Also
        --------
        `~greykite.common.viz.timeseries_plotting.add_groupby_column` : called by this function
        `~greykite.common.viz.timeseries_plotting.flexible_grouping_evaluation` : called by this function
        """
        df = self.df_train if which.lower() == "train" else self.df_test
        df = df.copy()
        result = add_groupby_column(
            df=df,
            time_col=self.time_col,
            groupby_time_feature=groupby_time_feature,
            groupby_sliding_window_size=groupby_sliding_window_size,
            groupby_custom_column=groupby_custom_column)
        df = result["df"]

        map_func_dict = self.autocomplete_map_func_dict(map_func_dict)
        grouped_df = flexible_grouping_evaluation(
            df,
            map_func_dict=map_func_dict,
            groupby_col=result["groupby_col"],
            agg_kwargs=agg_kwargs,
            extend_col_names=extend_col_names)

        return grouped_df

    def plot_flexible_grouping_evaluation(
            self,
            which="train",
            groupby_time_feature=None,
            groupby_sliding_window_size=None,
            groupby_custom_column=None,
            map_func_dict=None,
            agg_kwargs=None,
            extend_col_names=False,
            y_col_style_dict="auto-fill",
            default_color="rgba(0, 145, 202, 1.0)",
            xlabel=None,
            ylabel=None,
            title=None,
            showlegend=True):
        """Plots group-wise evaluation metrics. Whereas
        `~greykite.framework.output.univariate_forecast.UnivariateForecast.plot_grouping_evaluation`
        shows one metric, this can show any number of custom metrics.

        For example:

            * Mean and quantiles of squared error by group.
            * Mean and quantiles of residuals by group.
            * Mean and quantiles of actual and forecast by group.
            * % of actuals outside prediction intervals by group
            * any combination of the above metrics by the same group

        See `~greykite.framework.output.univariate_forecast.UnivariateForecast.get_flexible_grouping_evaluation`
        for details.

        which: `str`
            "train" or "test". Which dataset to evaluate.
        groupby_time_feature : `str` or None, optional
            If provided, groups by a column generated by
            `~greykite.common.features.timeseries_features.build_time_features_df`.
            See that function for valid values.
        groupby_sliding_window_size : `int` or None, optional
            If provided, sequentially partitions data into groups of size
            ``groupby_sliding_window_size``.
        groupby_custom_column : `pandas.Series` or None, optional
            If provided, groups by this column value. Should be same length as the DataFrame.
        map_func_dict : `dict` [`str`, `callable`] or None, default None
            Grouping evaluation metric specification, along with ``agg_kwargs``.
            See `~greykite.framework.output.univariate_forecast.UnivariateForecast.get_flexible_grouping_evaluation`.
        agg_kwargs : `dict` or None, default None
            Grouping evaluation metric specification, along with ``map_func_dict``.
            See `~greykite.framework.output.univariate_forecast.UnivariateForecast.get_flexible_grouping_evaluation`.
        extend_col_names : `bool` or None, default False
            How to name the grouping metrics.
            See `~greykite.framework.output.univariate_forecast.UnivariateForecast.get_flexible_grouping_evaluation`.
        y_col_style_dict: `dict` [`str`, `dict` or None] or "plotly" or "auto" or "auto-fill", default "auto-fill"
            The column(s) to plot on the y-axis, and how to style them. The names should match
            those generated by ``agg_kwargs`` and ``extend_col_names``.
            The function
            `~greykite.framework.output.univariate_forecast.UnivariateForecast.get_flexible_grouping_evaluation`
            can be used to check the column names.

            For convenience, start with "auto-fill" or "plotly", then adjust styling as needed.

            See `~greykite.common.viz.timeseries_plotting.plot_multivariate` for details.

        default_color: `str`, default "rgba(0, 145, 202, 1.0)" (blue)
            Default line color when ``y_col_style_dict`` is one of "auto", "auto-fill".
        xlabel : `str` or None, default None
            x-axis label. If None, default is ``x_col``.
        ylabel : `str` or None, default None
            y-axis label. If None, y-axis is not labeled.
        title : `str` or None, default None
            Plot title. If None and ``ylabel`` is provided, a default title is used.
        showlegend : `bool`, default True
            Whether to show the legend.

        Returns
        -------
        fig : `plotly.graph_objs.Figure`
            Interactive plotly graph showing the evaluation metrics.

            See `~greykite.common.viz.timeseries_plotting.plot_forecast_vs_actual`
            return value for how to plot the figure and add customization.

        See Also
        --------
        `~greykite.framework.output.univariate_forecast.UnivariateForecast.get_flexible_grouping_evaluation` : called by this function
        `~greykite.common.viz.timeseries_plotting.plot_multivariate` : called by this function
        """
        grouped_df = self.get_flexible_grouping_evaluation(
            which=which,
            groupby_time_feature=groupby_time_feature,
            groupby_sliding_window_size=groupby_sliding_window_size,
            groupby_custom_column=groupby_custom_column,
            map_func_dict=map_func_dict,
            agg_kwargs=agg_kwargs,
            extend_col_names=extend_col_names)

        x_col = grouped_df.index.name
        grouped_df.reset_index(inplace=True)
        fig = plot_multivariate(
            grouped_df,
            x_col=x_col,
            y_col_style_dict=y_col_style_dict,
            default_color=default_color,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            showlegend=showlegend)
        return fig

    def make_univariate_time_series(self):
        """Converts prediction into a UnivariateTimeSeries
        Useful to convert a forecast into the input regressor for a subsequent forecast.

        :return: UnivariateTimeSeries
        """
        ts = UnivariateTimeSeries()
        df = (self.df[[self.time_col, self.predicted_col]]
              .rename({self.predicted_col: self.ylabel}, axis=1))
        ts.load_data(df, self.time_col, self.ylabel)
        return ts

    def plot_components(self, **kwargs):
        """Class method to plot the components of a `UnivariateForecast` object.

        ``Silverkite`` calculates component plots based on ``fit`` dataset.
        ``Prophet`` calculates component plots based on ``predict`` dataset.

        For estimator specific component plots with advanced plotting options call
        ``self.estimator.plot_components()``.

        Returns
        -------
        fig: `plotly.graph_objs.Figure` for ``Silverkite``
             `matplotlib.figure.Figure` for ``Prophet``
             Figure plotting components against appropriate time scale.
        """
        return self.estimator.plot_components(**kwargs)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Convex optimization approach to forecast reconciliation."""

import functools
import logging
import sys
import warnings

import cvxpy as cp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.linalg import sqrtm
from sklearn.exceptions import NotFittedError

from greykite.algo.reconcile.hierarchical_relationship import HierarchicalRelationship
from greykite.common.python_utils import reorder_columns


try:
    from IPython.core import display as ICD
except ImportError:
    pass  # ipython is an optional dependency, by default only enabled for development


def get_weight_matrix(weights, n_forecasts, name, weight_auto=None):
    """Returns a diagonal weight matrix with shape (``n_forecasts``, ``n_forecasts``).

    Parameters
    ----------
    weights : `list` [`float`] or "auto" or None
        What weights to use.

            - If a list, returns a diagonal matrix with the list values on the diagonal.
              These values specify the weight for each timeseries.
              In ``ReconcileAdditiveForecasts``, weights are applied to the matrix
              whose rows are reordered to canonical form (transposed output of ``reorder_columns``)
            - If "auto", determined by ``weight_auto``.
            - If None, the identity matrix (equal weights).

    n_forecasts : `int`
        The number of forecasts (shape of the output).
    name : `str`
        The name of the weight matrix.
    weight_auto : `numpy.array` or None
        Weights to use if ``array="auto"``.
        Should be a square matrix with shape (``n_forecasts``, ``n_forecasts``).
        None corresponds to the identity matrix.

    Returns
    -------
    weight_matrix : `numpy.array`
        Weights to apply to the errors before taking the norm.
        Diagonal matrix with shape (``n_forecasts``, ``n_forecasts``)
    """
    if weights is None:
        weight_matrix = np.eye(n_forecasts)
    elif isinstance(weights, str) and weights == "auto":
        if weight_auto is None:
            logging.info(f"'auto' weight for {name} is the identity matrix")
            weight_matrix = np.eye(n_forecasts)
        else:
            logging.info(f"'auto' weight for {name} is {weight_auto}")
            weight_matrix = weight_auto
    else:
        weight_matrix = np.diag(weights)

    if weight_matrix.shape != (n_forecasts, n_forecasts):
        raise ValueError(
            f"Expected square matrix with size {n_forecasts}, but `{name}` has "
            f"weight matrix with shape {weight_matrix.shape}")
    return weight_matrix


def get_fit_params(method=None):
    """Returns parameters for
    `greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts.fit`
    corresponding to recognized hierarchical reconciliation method.

    Parameters
    ----------
    method : `str` or None
        Which reconciliation method to use
        Valid values are "bottom_up", "ols", "mint_sample".

            - "bottom_up"   : Sums leaf nodes. Unbiased transform that uses only the values of the leaf nodes
                              to propagate up the tree. Each node's value is the sum of its
                              corresponding leaf nodes' values (a leaf node corresponds to a node T if it is
                              a leaf node of the subtree with T as its root, i.e. a descendant of T or T itself).
                              See Dangerfield and Morris 1992 "Top-down or bottom-up: Aggregate versus disaggregate
                              extrapolations" for one discussion of this method.
            - "ols"         : OLS estimate proposed by https://robjhyndman.com/papers/Hierarchical6.pdf
                              (Hyndman et al. 2010, "Optimal combination forecasts for hierarchical time series")
                              Also see https://robjhyndman.com/papers/mint.pdf section 2.4.1.
                              (Wickramasuriya et al. 2019 "Optimal forecast reconciliation for
                              hierarchical and grouped time series through trace minimization".)
                              Unbiased transform that minimizes variance of adjusted residuals,
                              using "identity" estimate of original residual variance.
                              Optimal if original forecast errors are uncorrelated with equal variance (unlikely).
                              Depends only on the structure of the hierarchy, not on the data itself.
            - "mint_sample" : Unbiased transform that minimizes variance of adjusted residuals,
                              using "sample" estimate of original residual variance.
                              See Wickramasuriya et al. 2019 section 2.4.4.
                              Depends only on the structure of the hierarchy and forecast error covariances,
                              not on the data itself.

        If None, returns an empty dictionary.

    Returns
    -------
    estimator_fit_params : `dict` [`str`, Any]
        Parameters to use when calling ``ReconcileAdditiveForecasts.fit()``.
    """
    if method is None:
        estimator_fit_params = {}
    elif method == "bottom_up":
        logging.info("Using bottom_up estimator")
        estimator_fit_params = {}  # already passed to ``fit`` via ``method``
    elif method == "ols":
        logging.info("Using ols estimator")
        estimator_fit_params = dict(
            lower_bound=None,
            upper_bound=None,
            unbiased=True,          # unbiased
            lam_adj=0.0,
            lam_bias=0.0,
            lam_coef=0.0,
            lam_train=0.0,
            lam_var=1.0,            # minimizes variance
            covariance="identity",  # assumes equal uncorrelated variance
            weight_adj=None,
            weight_bias=None,
            weight_coef=None,
            weight_train=None,
            weight_var=None,
        )
    elif method == "mint_sample":
        logging.info("Using mint_sample estimator")
        estimator_fit_params = dict(
            lower_bound=None,
            upper_bound=None,
            unbiased=True,        # unbiased
            lam_adj=0.0,
            lam_bias=0.0,
            lam_coef=0.0,
            lam_train=0.0,
            lam_var=1.0,          # minimizes variance
            covariance="sample",  # in-sample variance estimate
            weight_adj=None,
            weight_bias=None,
            weight_coef=None,
            weight_train=None,
            weight_var=None,
        )
    else:
        raise ValueError(f"`method` '{method}' is not recognized. "
                         f"Must be one of 'bottom_up', 'ols', 'mint_sample' or None")
    return estimator_fit_params


def apply_method_defaults(fit_func):
    """Decorator for `greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts.fit`.
    Fetches parameters based on ``method`` and calls ``fit_func`` with the result.

    Parameters
    ----------
    fit_func : `callable`
        Should be `greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts.fit`.

    Returns
    -------
    apply_defaults_and_fit : callable
        Function that overwrites ``fit_params`` with those specified by ``method``
        and calls ``fit_func`` with those params.
    """
    @functools.wraps(fit_func)
    def apply_defaults_and_fit(
            self,
            forecasts,
            actuals,
            **fit_params):
        """Overwrites ``fit_params`` with those specified by ``fit_params["method"]``.
        Calls ``fit_func`` with those params.

        Parameters
        ----------
        self : `ReconcileAdditiveForecasts`
            Self. Decorator should be applied to ``ReconcileAdditiveForecasts.fit()``.
        forecasts : `pandas.DataFrame`
            Forecasted values to pass to ``fit_func``.
        actuals : `pandas.DataFrame`
            Actual values to pass to ``fit_func``.
        fit_params : `dict`
            Original parameters used to call ``fit_func``.
            These are overridden by the ones corresponding to ``method``.

        Returns
        -------
        output : Any
            Same as the output of ``fit_func``.
        """
        method = fit_params.get("method")
        method_fit_params = get_fit_params(method)
        params = dict(fit_params, **method_fit_params)
        return fit_func(
            self,
            forecasts,
            actuals,
            **params)
    return apply_defaults_and_fit


class ReconcileAdditiveForecasts:
    """Reconciles forecasts to satisfy additive constraints.

    Constraints can be encoded by the tree structure via ``levels``.
    In the tree formulation, a parent's value must be the sum of its children's values.

    Or, constraints can be encoded as a matrix via ``constraint_matrix``,
    specifying additive equations that must sum to 0. The constraints need not have a
    tree representation.

    Provides standard methods such as bottom up, ols, MinT. Also supports a custom method
    that minimizes user-specified types of error. The solution is derived by convex optimization.
    If desired, a constraint is added to require the transformation to be unbiased.

    If not using method="ols" or method="bottom_up", which don't depend on the data,
    forecast reconciliation should be trained once per horizon
    (# periods between forecasted date and train_end_date),
    because the optimal adjustment may differ.

    See http://go/raf for details.

    Attributes
    ----------
    forecasts : `pandas.DataFrame`, shape (n, m)
        Original forecasted values, used to train the method.
        Also known as "base" forecasts.
        Long format where each column is a time series.
        and each row is a time step.
        For proper variance estimates for the variance penalty,
        values should be at a fixed-horizon (e.g. always 7-step ahead).
    actuals : `pandas.DataFrame`, shape (n, m)
        Actual values to train the method, corresponding to ``forecasts``.
        Must have the same shape and column names as ``forecasts``.
    constraint_matrix : `numpy.array`, shape (c, m), or None
        Constraints. ``c x m`` array encoding ``c`` constraints of ``m`` variables.
        We require ``constraint_matrix @ transform_matrix = 0``.
        For example, to encode ``-x1 + x2 + x3 == 0 and -x2 + x4 + x5 == 0``::

            constraint_matrix = np.array([
                [-1, 1, 1, 0, 0],
                [0, -1, 0, 1, 1]])

        Entries are typically in [-1, 0, 1], but this is not required.
        Either ``constraint_matrix`` or ``levels`` must be provided.

    levels : `list` [`list` [`int`]] or None
        A simpler way to encode tree constraints. Overrides ``constraint_matrix`` if provided.
        Specifies the number of children of each parent (internal) node in the tree.
        The number of inner lists is the height of the tree. The ith inner list provides the number
        of children of each node at depth i. For example::

            # root node with 3 children
            levels = [[3]]
            # root node with 3 children, who have 2, 3, 3 children respectively
            levels = [[3], [2, 3, 3]]

        All leaf nodes must have the same depth. Thus, the first sublist must have one
        integer, the length of a sublist must equal the sum of the previous sublist,
        and all integers in ``levels`` must be positive.

        Either ``constraint_matrix`` or ``levels`` must be provided.

    order_dict : `dict` [`str`, `float`] or None
        How to order the columns before fitting.
        The key is the column name, the value is its position.
        When ``levels`` is used, map each column name to the
        order of its corresponding node in a BFS traversal of the tree.
        When ``constraint_matrix`` is used, this shuffles the order
        of the columns before the constraints are applied (thus, columns
        in ``constraint_matrix`` refer to the columns after reordering).

        If None, no reordering is done.

    method : `str` or None
        Which method to use.
        If None, uses the parameters specified to formulate the convex optimization problem.
        If a string, valid values are "bottom_up", "ols", "mint_sample". The other parameters are ignored.

            - "bottom_up"   : Sums leaf nodes. Unbiased transform that uses only the values of the leaf nodes
                              to propagate up the tree. Each node's value is the sum of its
                              corresponding leaf nodes' values (a leaf node corresponds to a node T if it is
                              a leaf node of the subtree with T as its root, i.e. a descendant of T or T itself).
                              See Dangerfield and Morris 1992 "Top-down or bottom-up: Aggregate versus disaggregate
                              extrapolations" for one discussion of this method.
            - "ols"         : OLS estimate proposed by https://robjhyndman.com/papers/Hierarchical6.pdf
                              (Hyndman et al. 2010, "Optimal combination forecasts for hierarchical time series")
                              Also see https://robjhyndman.com/papers/mint.pdf section 2.4.1.
                              (Wickramasuriya et al. 2019 "Optimal forecast reconciliation for
                              hierarchical and grouped time series through trace minimization".)
                              Unbiased transform that minimizes variance of adjusted residuals,
                              using "identity" estimate of original residual variance.
                              Optimal if original forecast errors are uncorrelated with equal variance (unlikely).
                              Depends only on the structure of the hierarchy, not on the data itself.
            - "mint_sample" : Unbiased transform that minimizes variance of adjusted residuals,
                              using "sample" estimate of original residual variance.
                              See Wickramasuriya et al. 2019 section 2.4.4.
                              Depends only on the structure of the hierarchy and forecast error covariances,
                              not on the data itself.

    lower_bound : `float` or None
        Lower bound on each entry of ``transform_matrix``.
        If None, no lower bound is applied.
    upper_bound : `float` or None
        Upper bound on each entry of ``transform_matrix``.
        If None, no upper bound is applied.
    unbiased : `bool`
        Whether the resulting transformation must be unbiased.
    lam_adj : `float`
        Weight for the adjustment penalty.
        The adjustment penalty is the mean squared difference
        between adjusted forecasts and base forecasts.
    lam_bias : `float`
        Weight for the bias penalty.
        The bias penalty is the mean squared difference
        between adjusted actuals and actuals.
        The value is 0 for an unbiased transformation,
        actuals satisfy the constraints.
    lam_coef : `float`
        Weight for the coefficient penalty.
        The coefficient penalty measures the deviation of the
        transformation matrix from the identity matrix.
        This penalty is more ad-hoc, and its units are not on
        the same scale as the other. The unweighted penalty tends
        to be large, so prefer smaller values of ``lam_coef`` relative
        to other penalties, if desired.
    lam_train : `float`
        Weight for the training MSE penalty.
        The train MSE penalty measures the mean squared difference
        between adjusted forecasts and actuals.
    lam_var : `float`
        Weight for the variance penalty.
        The variance penalty measures the variance of
        adjusted forecast errors an unbiased transformation.
        It is reported as the average of the variances across timeseries.
        It is based on the variance of the base forecast error
        variance, ``covariance``. For biased transforms,
        this is an underestimate of the true variance.
    covariance : `numpy.array` of shape (m, m), or "sample" or "identity"
        Variance-covariance matrix of base forecast errors. Used to
        compute the variance penalty.

            - If a `numpy.array`, row/column i corresponds to the
              ith column after reordering by ``order_dict``. Should
              be reported on the original scale of the data.
            - If "sample", the sample covariance of residuals.
            - If "identity", the identity matrix.

    weight_adj : `list` [`float`] of length m, or "auto" or None
        Weight for the adjustment penalty that allows a different
        weight per-timeseries.

            - If a list, values specify the weight for each forecast
              after reordering by ``order_dict``.
            - If "auto", proportional to the MedAPE of the forecast.
            - If None, the identity matrix (equal weights).

    weight_bias : `list` [`float`] of length m, or "auto" or None
        Weight for the bias penalty that allows a different
        weight per-timeseries.

            - If a list, values specify the weight for each forecast
              after reordering by ``order_dict``.
            - If "auto", proportional to the MedAPE of the forecast.
            - If None, the identity matrix (equal weights).

    weight_coef : `numpy.array` of shape (m, m), or "auto" or None
        Weight for the coefficient penalty that allows a different
        weight per-timeseries.

            - If a `numpy.array`, values specify the weight for each entry in
              ``transform_matrix - np.eye(m)``.
              Note that rows/cols in ``transform_matrix``
              correspond to the forecasts after reordering by ``order_dict``.
            - If "auto", a matrix of 1s, with 0s on the diagonal.
            - If None, a matrix of 1s.

    weight_train : `list` [`float`] of length m, or None
        Weight for the train MSE penalty that allows a different
        weight per-timeseries.

            - If a list, values specify the weight for each forecast
              after reordering by ``order_dict``.
            - If None, the identity matrix (equal weights).

    weight_var : `list` [`float`] of length m, or None
        Weight for the variance penalty that allows a different
        weight per-timeseries.

            - If a list, values specify the weight for each forecast
              after reordering by ``order_dict``.
            - If None, the identity matrix (equal weights).

    names : `pandas.Index`
        Names of ``forecast`` columns after reordering by ``order_dict``.
    tree : `greykite.algo.reconcile.hierarchical_relationship.HierarchicalRelationship` or None
        If ``levels`` is provided, represents the tree structure encoded by the levels.
        Else None.
    transform_variable : `cvxpy.Variable`, shape (m, m) or None
        Optimization variable to learn the transform matrix.
        None if a rule-based method is used, e.g. ``method == bottom_up``
    transform_matrix : `numpy.array`, shape (m, m)
        Transformation matrix. Same as ``transform_variable.value``, unless
        the solver failed the find a solution, and a backup value is used.
        Adjusted forecasts are computed by applying the transform from the left
        to reordered and transposed ``forecasts``. See ``transform`` in this class.
    prob : `cvxpy.Problem`
        Convex optimization problem.
    is_optimization_solution : `bool`
        Whether ``transform_matrix`` is a solution found by convex optimization solution.
        If False, then ``transform_matrix`` may be set to a backup value (bottom up transform).
        Check ``prob.status`` for more details about solver status.
    objective_fn : `callable`
        Evaluates the objective function for a given transform matrix and dataset.
        Takes ``transform_matrix, forecast_matrix (optional), actual_matrix (optional)``.
        Return value has same format as ``objective_fn_val``.
        If forecast_matrix/actual_matrix are not provided, uses the fitting datasets.
    objective_fn_val : `dict` [`str`, `float`]
        Dictionary containing the objective value, and its components,
        as evaluated on the training set for the identified optimal solution
        from convex optimization.
        Keys are:

            ``"adj"``    : adjustment size
            ``"bias"``   : bias of estimator
            ``"coef"``   : penalty on off-diagonal coefficients
            ``"train"``  : train set MSE
            ``"var"``    : variance of unbiased estimator
            ``"total"``  : sum of the above

    objective_weights : `dict` [`str`, `np.array` of shape (m, m)]
        Weights used in the objective function, derived from
        ``covariance``, ``weight_*``, ``forecasts``, ``actuals``.
        Keys are:

            - weight_adj
            - weight_bias
            - weight_coef
            - weight_train
            - weight_var
            - covariance

    adjusted_forecasts : `pandas.DataFrame`, shape (n, m)
        Adjusted ``forecasts`` that satisfy the constraints.
    constraint_violation : `dict` [`str`, `float`]
        The normalized constraint violations on training set.
        Keys are "actual", "forecast", and "adjusted".
        Root mean squared constraint violation is divided by root mean squared actual value.
    evaluation_df : `pandas.DataFrame`, shape (m, # metrics)
        DataFrame of evaluation results on training set. Rows are timeseries,
        columns are metrics. See ``evaluate`` in this class.
    forecasts_test : `pandas.DataFrame`, shape (q, m)
        Forecasted values to test the method.
        Long format where each column is a time series
        and each row is a time step.
        Must have the same column names as ``forecasts``.
        Can have a different number of rows (observations).
    actuals_test : `pandas.DataFrame`, shape (q, m)
        Actual values to test the method.
        Must have the same shape and column names as ``forecasts_test``.
    adjusted_forecasts_test : `pandas.DataFrame`, shape (q, m)
        Adjusted ``forecasts_test`` that satisfy the constraints.
    constraint_violation_test : `dict` [`str`, `float`]
        The normalized constraint violations on test set.
        Keys are "actual", "forecast", and "adjusted".
        Root mean squared constraint violation is divided by root mean squared actual value
        on test set.
    evaluation_df_test : `pandas.DataFrame`, shape (m, # metrics)
        DataFrame of evaluation results on test set. Rows are timeseries,
        columns are metrics. See ``evaluate()`` in this class.

    Methods
    -------
    fit : callable
        Fits the ``transform_matrix`` from training data.
    transform : callable
        Adjusts a forecast to satisfy additive constraints using the ``transform_matrix``.
    evaluate : callable
        Evaluates the adjustment quality by its impact to MAPE, MedAPE, and RMSE.
    fit_transform : callable
        Fits and transforms the training data.
    fit_transform_evaluate : callable
        Fits, transforms, and evaluates on training data.
    transform_evaluate : calllable
        Transforms and evaluates on a new test set.
    """
    def __init__(self):
        # training data and input params
        self.forecasts = None
        self.actuals = None
        self.constraint_matrix = None  # updated by `fit` if `levels` is provided
        self.levels = None
        self.order_dict = None
        self.method = None
        self.lower_bound = None
        self.upper_bound = None
        self.unbiased = False
        self.lam_adj = 0.0
        self.lam_bias = 0.0
        self.lam_coef = 0.0
        self.lam_train = 0.0
        self.lam_var = 0.0
        self.covariance = None
        self.weight_adj = None
        self.weight_bias = None
        self.weight_coef = None
        self.weight_train = None
        self.weight_var = None

        # set by `fit`
        self.names = None
        self.tree = None
        self.transform_variable = None
        self.transform_matrix = None
        self.prob = None
        self.is_optimization_solution = None
        self.objective_fn = None
        self.objective_fn_val = None
        self.objective_weights = None

        # training data transformed result
        self.adjusted_forecasts = None
        self.constraint_violation = None
        self.evaluation_df = None
        # test data transformed result
        self.forecasts_test = None
        self.actuals_test = None
        self.adjusted_forecasts_test = None
        self.constraint_violation_test = None
        self.evaluation_df_test = None

    def _form_constraints(
            self,
            transform_variable,
            Ya=None):
        """Forms the constraints for convex optimization.
        The following attributes should be set before calling this method:

            - constraint_matrix or tree (required)
            - unbiased                  (optional)
            - lower_bound               (optional)
            - upper_bound               (optional)

        Parameters
        ----------
        transform_variable : `cvxpy.Variable`, shape (m, m)
            Optimization variable to learn the transform matrix.
        Ya : `numpy.array`, shape (m, n) or None
            Actuals in wide format. Rows are timeseries, columns are timestamps.
            Must have the same shape as ``Yf``.

            Created in ``fit`` by reordering the columns of ``self.actuals``
            according to ``self.order_dict``, taking the transpose,
            and scaling to mean 1.0.

        Returns
        -------
        constraints : `list` [`cvxpy.constraints.constraint.Constraint`]
            Constraints for the convex optimization problem.
            Requires additive consistency as specified by ``self.constraint_matrix``.
            If ``self.unbiased``, requires unbiasedness.
        """
        constraints = []
        if self.lower_bound is not None:
            constraints += [transform_variable >= self.lower_bound]
        if self.upper_bound is not None:
            constraints += [transform_variable <= self.upper_bound]

        if self.tree is not None:
            # Uses tree's constraints if `tree` is provided
            constraint_matrix = self.tree.constraint_matrix
        else:
            constraint_matrix = self.constraint_matrix

        # If this holds, then the constraints are satisfied for any Yf, i.e.
        # constraint_matrix @ Y_{adjusted} = constraint_matrix @ transform_variable @ Yf == 0.
        constraints += [constraint_matrix @ transform_variable == 0]

        if self.unbiased:
            if self.tree is None:
                if Ya is None:
                    raise ValueError("`Ya` must be provided if `unbiased` and `tree` is None.")
                # Since we don't know the structure, we use an
                # empirical check for unbiasedness.
                constraints += [transform_variable @ Ya == Ya]
            else:
                # Let there be `ell` leaf forecasts.
                # S is m x ell, summing matrix against leaf nodes. Y_{actual} = S @ Y_{leaf}
                # See sec. 2.1, Wickramasuriya et al. 2019 "Optimal forecast reconciliation
                # for hierarchical and grouped time series through trace minimization".
                # transform_variable T is defined as S @ P. T is unbiased if T @ S = S @ P @ S = S.
                # (i.e. Forecasts that satisfy hierarchical constraints remain consistent)
                sum_matrix = self.tree.sum_matrix
                constraints += [transform_variable @ sum_matrix == sum_matrix]

        if Ya is not None:
            violation_magnitude = np.sqrt(np.mean(np.square(constraint_matrix @ Ya)))
            observation_magnitude = np.sqrt(np.mean(np.square(Ya)))
            relative_error = violation_magnitude / observation_magnitude
            if relative_error > 1e-3:
                warnings.warn(f"Actuals do not satisfy the constraints! Relative error is {relative_error.round(5)}")
        return constraints

    def _form_objective(
            self,
            Yf,
            Ya,
            transform_variable,
            covariance=None):
        """Forms the objective for convex optimization.

        The objective is a sum of five types of errors::

            obj = adj + bias + coef + train + var

        Where

            ``"adj"`` : MSE of adjusted forecasts vs base forecasts (adjustment size)
            ``"bias"`` : MSE of adjusted actuals vs actuals (squared bias)
            ``"coef"`` : deviation of transformation from identity matrix
            ``"train"`` : MSE of adjusted forecasts vs actuals (training error)
            ``"var"`` : estimated variance of adjusted forecast residuals (variance)

        Relative importance of these errors is specified by ``lam_*``.
        Relative importance of forecasts / coefficients is specified by ``weight_*``.

        The following attributes should be set before calling this method:

            - lam_adj       (optional)
            - lam_bias      (optional)
            - lam_coef      (optional)
            - lam_train     (optional)
            - lam_var       (optional)
            - weight_adj    (optional)
            - weight_bias   (optional)
            - weight_coef   (optional)
            - weight_train  (optional)
            - weight_var    (optional)

        Parameters
        ----------
        Yf : `numpy.array`, shape (m, n)
             Forecasts in wide format. Rows are timeseries, columns are timestamps.
             The goal is to adjust these values to satisfy the constraints.

             Created in ``fit`` by reordering the columns of ``self.forecasts``
             according to ``self.order_dict`` and taking the transpose,
             and scaling down by the mean of the actuals.
        Ya : `numpy.array`, shape (m, n)
             Actuals in wide format. Rows are timeseries, columns are timestamps.
             Must have the same shape as ``Yf``.

             Created in ``fit`` by reordering the columns of ``self.actuals``
             according to ``self.order_dict``, taking the transpose,
             and scaling to mean 1.0.
        transform_variable : `cvxpy.Variable`, shape (m, m)
            Optimization variable. Square matrix with size matching the rows in ``Yf``.

        Returns
        -------
        obj : `cvxpy.problems.objective.Minimize`
            Objective function to minimize.

        Updates ``self.objective_weights`` to contain the
        matrices used in the objective.
        """
        n_forecasts = Yf.shape[0]
        if (isinstance(self.weight_adj, str) and self.weight_adj == "auto") or \
                (isinstance(self.weight_bias, str) and self.weight_bias == "auto"):
            # Default weight is the Med-APE for each timeseries
            weight_auto = np.diag(np.nanmedian(np.abs(Ya - Yf) / np.abs(Ya), axis=1))
            # Rescales the weight so it has the same norm as `m x m` identity matrix
            weight_auto = weight_auto * np.linalg.norm(np.eye(n_forecasts)) / np.linalg.norm(weight_auto)
        else:
            weight_auto = None
        weight_adj = get_weight_matrix(weights=self.weight_adj, n_forecasts=n_forecasts, name="weight_adj", weight_auto=weight_auto)
        weight_bias = get_weight_matrix(weights=self.weight_bias, n_forecasts=n_forecasts, name="weight_bias", weight_auto=weight_auto)
        if self.weight_coef is None:
            weight_coef = np.ones((n_forecasts, n_forecasts))
        elif isinstance(self.weight_coef, str) and self.weight_coef == "auto":
            weight_coef = np.ones((n_forecasts, n_forecasts)) - np.eye(n_forecasts)
        else:
            weight_coef = self.weight_coef
        weight_train = get_weight_matrix(weights=self.weight_train, n_forecasts=n_forecasts, name="weight_train")
        weight_var = get_weight_matrix(weights=self.weight_var, n_forecasts=n_forecasts, name="weight_var")

        if isinstance(covariance, str):
            if covariance == "identity":
                covariance = np.eye(n_forecasts)
            elif covariance == "sample":
                # The sample covariance matrix estimate of h-step ahead forecast errors.
                # Assumes the forecast horizon (h) is fixed.
                # Pass `covariance` to `fit` if a better estimate is available.
                residuals = Ya - Yf
                covariance = np.cov(residuals)
            else:
                raise ValueError("`covariance` not recognized. Provide a valid string in ['identity', 'sample'] or a matrix.")

        self.objective_weights = {
            "weight_adj": weight_adj,
            "weight_bias": weight_bias,
            "weight_coef": weight_coef,
            "weight_train": weight_train,
            "weight_var": weight_var,
            "covariance": covariance
        }

        # Mean squared adjustment size (on observed forecasts distribution, weighted by ``weight_adj``)
        err_adj = cp.sum_squares(cp.norm(weight_adj @ (transform_variable @ Yf - Yf), p="fro")) / np.size(Yf)
        # Mean squared bias (on observed actuals distribution, weighted by ``weight_bias``)
        err_bias = cp.sum_squares(cp.norm(weight_bias @ (transform_variable @ Ya - Ya), p="fro")) / np.size(Ya)
        # Mean squared deviation from identity matrix
        err_coef = cp.sum_squares(cp.norm(cp.multiply(weight_coef, transform_variable - np.eye(n_forecasts)), p="fro")) / np.size(weight_coef)
        # Mean squared train error (proxy for bias^2 + variance) (on observed distribution, weighted by ``weight_train``)
        err_train = cp.sum_squares(cp.norm(weight_train @ (transform_variable @ Yf - Ya), p="fro")) / np.size(Yf)
        # Variance of an unbiased estimator.
        # We can't optimize the trace directly, as specified by Wickramasuriya et al. 2019.
        #   err_var = cp.trace(weight_var @ transform_variable @ covariance @ transform_variable.T) / n_forecasts if covariance is not None else 0.0
        #   "Problem does not follow DCP rules. Specifically:"
        #   "The objective is not DCP."
        # Because `covariance` is symmetric and PSD, we can rewrite the trace as a squared Frobenius norm.
        #   tr(T covariance T') = ||T sqrt(covariance)||_{F]^{2}
        # Note that `weight_var` has a different interpretation than when
        # inside the trace. It matches the interpretation of the other weights.
        # Divides by `n_forecasts` for mean adjusted forecast error variance for a timeseries
        #   (the trace is the sum across timeseries).
        err_var = cp.sum_squares(cp.norm(weight_var @ transform_variable @ sqrtm(covariance), p="fro")) / n_forecasts if covariance is not None else 0.0

        lams = [self.lam_adj, self.lam_bias, self.lam_coef, self.lam_train, self.lam_var]
        errs = [err_adj, err_bias, err_coef, err_train, err_var]
        obj = cp.Minimize(cp.scalar_product(lams, errs))

        def objective_fn(transform_matrix, forecast_matrix=Yf, actual_matrix=Ya):
            """Evaluates the objective function for a given transform matrix and dataset.

            Parameters
            ----------
            transform_matrix : `numpy.array`, shape (m, m)
                Applied to forecast and actual matrices from the left to get the transformed result.
            forecast_matrix : `numpy.array`, shape (m, n), optional
                The forecasts in wide array format.
                Must have the same shape as ``actual_matrix``.
                If not provided, uses the forecasts for fitting.
            actual_matrix : `numpy.array`, shape (m, n), optional
                The actuals in wide array format.
                Must have the same number of rows as ``transform_matrix``.
                If not provided, uses the actuals for fitting.

            Returns
            -------
            objective_value : `dict` [`str`, `float`] or None
                Dictionary containing the objective value, and its components.
                Keys are:

                    ``"adj"``    : adjustment size
                    ``"bias"``   : bias of estimator
                    ``"coef"``   : penalty on off-diagonal coefficients
                    ``"train"``  : train set MSE
                    ``"var"``    : variance of unbiased estimator
                    ``"total"``  : sum of the above

                Returns None if ``transform_matrix`` is None.
            """
            if transform_matrix is None:
                return None
            # NB: Frobenius norm is the default norm
            np_err_adj = np.linalg.norm(weight_adj @ (transform_matrix @ forecast_matrix - forecast_matrix))**2 / np.size(forecast_matrix)
            np_err_bias = np.linalg.norm(weight_bias @ (transform_matrix @ actual_matrix - actual_matrix))**2 / np.size(actual_matrix)
            np_err_coef = np.linalg.norm(np.multiply(weight_coef, transform_matrix - np.eye(n_forecasts)))**2 / np.size(weight_coef)
            np_err_train = np.linalg.norm(weight_train @ (transform_matrix @ forecast_matrix - actual_matrix))**2 / np.size(actual_matrix)
            np_err_var = np.linalg.norm(weight_var @ transform_matrix @ sqrtm(covariance))**2 / n_forecasts if covariance is not None else 0.0
            np_errs = [np_err_adj, np_err_bias, np_err_coef, np_err_train, np_err_var]
            return {
                "adj": self.lam_adj * np_err_adj,        # adjustment size
                "bias": self.lam_bias * np_err_bias,     # bias^2 of estimator
                "coef": self.lam_coef * np_err_coef,     # penalty on off-diagonal coefficients
                "train": self.lam_train * np_err_train,  # train error (bias^2 + variance)
                "var": self.lam_var * np_err_var,        # variance of an *unbiased* estimator
                "total": np.dot(lams, np_errs)           # Sum of the above
            }
        self.objective_fn = objective_fn

        return obj

    @apply_method_defaults
    def fit(self,
            forecasts,
            actuals,
            order_dict=None,
            method=None,
            levels=None,
            constraint_matrix=None,
            lower_bound=None,
            upper_bound=None,
            unbiased=False,
            lam_adj=0.0,
            lam_bias=0.0,
            lam_coef=0.0,
            lam_train=0.0,
            lam_var=0.0,
            covariance="sample",
            weight_adj="auto",
            weight_bias="auto",
            weight_coef="auto",
            weight_train=None,
            weight_var=None,
            **solver_kwargs):
        """Fits the ``transform_matrix`` based on input data, constraint, and objective function.

        Sets the attributes between ``forecasts`` and ``objective_weights`` as noted
        in the class description, including ``transform_matrix``, ``transform_variable``,
        ``prob``, ``objective_fn_val``.

        If method != "bottom_up" and there is no solution, gives a warning and
        ``self.is_optimization_solution`` is set to False. Uses "bottom_up" solution
        as fallback approach if ``levels`` is provided.

        Parameters
        ----------
        forecasts : `pandas.DataFrame`, shape (n, m)
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        actuals : `pandas.DataFrame`, shape (n, m)
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        order_dict : `dict` [`str`, `float`] or None, default None
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        method : `str` or None, default None
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
            If provided, the parameters between ``lower_bound`` and ``weight_var`` below are ignored.
        levels : `list` [`list` [`int`]] or None, default None
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        constraint_matrix : `numpy.array`, shape (c, m) or None, default None
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        lower_bound : `float` or None, default None
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        upper_bound : `float` or None, default None
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        unbiased : `bool`, default True
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        lam_adj : `float`, default 0.0
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        lam_bias : `float`, default 0.0
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        lam_coef : `float`, default 0.0
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        lam_train : `float`, default 0.0
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        lam_var : `float`, default 0.0
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        covariance : `numpy.array` of shape (m, m), or "sample" or "identity", default "sample"
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        weight_adj : `list` [`float`] of length m, or "auto" or None, default "auto"
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        weight_bias : `list` [`float`] of length m, or "auto" or None, default "auto"
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        weight_coef : `numpy.array` of shape (m, m), or "auto" or None, default "auto"
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        weight_train : `list` [`float`] of length m, or None, default None
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        weight_var : `list` [`float`] of length m, or None, default None
            See attributes of `~greykite.algo.reconcile.convex.reconcile_forecasts.ReconcileAdditiveForecasts`.
        solver_kwargs : dict
            Specify the CVXPY solver and parameters. E.g. dict(verbose=True).
            See https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options.

        Returns
        -------
        transform_matrix : `numpy.array`, shape (m, m)
            Transformation matrix. Same as ``transform_variable.value``, unless
            the solver failed the find a solution, and a backup value is used.
            Adjusted forecasts are computed by applying the transform from the left
            to reordered and transposed ``forecasts``. See ``transform()`` in this class.
        """
        self.forecasts = reorder_columns(forecasts, order_dict=order_dict)  # Returns a copy
        self.actuals = reorder_columns(actuals, order_dict=order_dict)
        self.order_dict = order_dict
        self.levels = levels
        self.method = method
        self.constraint_matrix = constraint_matrix
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.unbiased = unbiased
        self.covariance = covariance
        self.lam_adj = lam_adj
        self.lam_bias = lam_bias
        self.lam_coef = lam_coef
        self.lam_train = lam_train
        self.lam_var = lam_var
        self.weight_adj = weight_adj
        self.weight_bias = weight_bias
        self.weight_coef = weight_coef
        self.weight_train = weight_train
        self.weight_var = weight_var

        if constraint_matrix is None and levels is None:
            raise ValueError("Either `constraint_matrix` or `levels` must be provided.")
        if constraint_matrix is not None and levels is not None:
            raise ValueError("Only one of `constraint_matrix` and `levels` can be provided.")
        if (method == "bottom_up") and levels is None:
            raise ValueError("Must provide `levels` if `method='bottom_up'`.")
        if lower_bound is not None and upper_bound is not None and lower_bound > upper_bound:
            raise ValueError(f"`lower_bound` {lower_bound} should not be greater than `upper_bound` {upper_bound}")
        if not unbiased and isinstance(covariance, str) and lam_var != 0.0:
            warnings.warn(f"Variance of residuals is underestimated if the estimator is biased.")

        if levels is not None:
            tree = HierarchicalRelationship(levels)
            self.tree = tree
            self.constraint_matrix = tree.constraint_matrix  # Overwrites ``constraint_matrix`` if ``levels`` is provided

        self.names = self.forecasts.columns
        if self.forecasts.shape != self.actuals.shape:
            raise ValueError(f"Forecast shape {self.forecasts.shape} does not match actuals shape {self.actuals.shape}")
        if self.constraint_matrix.shape[1] != self.forecasts.shape[1]:
            raise ValueError(f"The number of forecasts {self.forecasts.shape[1]} does not match "
                             f"the number of columns in `constraint_matrix` {self.constraint_matrix.shape[1]}. "
                             f"Make sure `levels` or `constraint_matrix` matches the data.")

        if method == "bottom_up":
            # Uses bottom-up method
            self.transform_matrix = self.tree.bottom_up_transform
        else:
            # Solves convex optimization problem to find transformation
            # Selects and reorders columns before matrix operations in constraint and objective
            Yf = np.array(self.forecasts).T  # rows are forecasts (wide-format)
            Ya = np.array(self.actuals).T

            # For numerical stability, scale units so that actuals have mean 1 before fitting.
            # The solution should be scale invariant.
            scale = Ya.mean()
            Yf /= scale
            Ya /= scale
            if covariance is not None and not isinstance(covariance, str):
                covariance = covariance / scale**2

            # Creates a matrix optimization variable.
            m = Yf.shape[0]
            transform_variable = cp.Variable((m, m))  # Linear transform matrix. Y_{adjusted} = transform_variable @ Yf

            # Creates constraints.
            constraints = self._form_constraints(
                transform_variable=transform_variable,
                Ya=Ya)

            # Forms objective.
            obj = self._form_objective(
                Yf=Yf,
                Ya=Ya,
                transform_variable=transform_variable,
                covariance=covariance)

            # Forms and solves problem.
            prob = cp.Problem(obj, constraints)
            try:
                prob.solve(**solver_kwargs)
            except cp.SolverError as e:
                warnings.warn(str(e))

            self.transform_variable = transform_variable
            self.transform_matrix = transform_variable.value
            self.prob = prob
            logging.info(f"Solver status: {prob.status}")
            logging.info(f"Solver optimal value: {prob.value}")
            logging.info(f"Adjustment matrix:\n {self.transform_matrix}")

            if self.transform_matrix is None:
                # No optimization solution is found.
                self.is_optimization_solution = False
                if self.tree is not None:
                    warnings.warn("Failed to find a solution. Falling back to bottom-up method.")
                    self.transform_matrix = self.tree.bottom_up_transform
                else:
                    warnings.warn("Failed to find a solution. Try setting CVXPY solver parameters, changing the "
                                  "constraints, or changing the objective weights")
            else:
                self.is_optimization_solution = True
            self.objective_fn_val = self.objective_fn(self.transform_matrix)
        return self.transform_matrix

    def transform(self, forecasts_test=None):
        """Transforms the provided forecasts using the
        fitted ``self.transform_matrix``.

        Parameters
        ----------
        forecasts_test : `pandas.DataFrame`, shape (r, m) or None
            Forecasted values to transform.
            Must have the same columns as ``self.forecasts``.
            If None, uses ``self.forecasts``.

        Returns
        -------
        adjusted_forecasts : `pandas.DataFrame`, shape (r, m)
            Adjusted forecasts that satisfy additive constraints.
            Columns are reordered according to ``self.order_dict``.

        If ``forecasts`` is None, results are stored to ``self.adjusted_forecasts``.
        Else, results are stored to ``self.adjusted_forecasts_test``, and the
        provided ``forecasts_test`` to ``self.forecasts_test``.
        """
        if self.transform_matrix is None:
            raise NotFittedError("Must call `fit` first.")

        is_train = forecasts_test is None
        if is_train:
            forecasts = self.forecasts
        else:
            forecasts = forecasts_test

        forecasts = reorder_columns(forecasts, order_dict=self.order_dict)  # Returns a copy
        Yf = np.array(forecasts).T  # Converts to wide format for transform
        adjusted_forecasts = pd.DataFrame(
            np.transpose(self.transform_matrix @ Yf),  # Converts back to long format
            columns=forecasts.columns,
            index=forecasts.index)

        if is_train:
            self.adjusted_forecasts = adjusted_forecasts
        else:
            self.forecasts_test = forecasts
            self.adjusted_forecasts_test = adjusted_forecasts

        return adjusted_forecasts

    def evaluate(self, is_train, actuals_test=None, plot=False, ipython_display=False):
        """Evaluates the adjustment quality. Computes the following metrics for each
        of the `m` timeseries:

             "Base MAPE"        : MAPE of base forecasts
             "Base MedAPE"      : MedAPE of base forecasts
             "Base RMSE"        : RMSE of base forecasts
             "Adjusted MAPE"    : MAPE of adjusted forecasts
             "Adjusted MedAPE"  : MedAPE of adjusted forecasts
             "Adjusted RMSE"    : RMSE of adjusted forecasts
             "RMSE % change"    : (Adjusted RMSE) / (Base RMSE) - 1
             "MAPE pp change"   : (Adjusted MAPE) - (Base MAPE)
             "MedAPE pp change" : (Adjusted MedAPE) - (Base MedAPE)

        Must call ``fit`` and ``transform`` before calling this method.

        Parameters
        ----------
        is_train : `bool`
            Whether to evaluate on training set or test set.
            If True, evaluates training adjustment quality.
            Else, evaluates test adjustment quality. In this
            case, ``actuals_test`` must be provided.
        actuals_test :  `pandas.DataFrame`
            Actual values on test set, required if ``is_train==False``.
            Must have the same shape as the forecasts passed to
            ``transform()``, i.e. ``self.forecasts_test.shape``.
        plot : `bool`, default False
            Whether to plot the results. Function is basic.
        ipython_display : `bool`, default False
            Whether to display the evaluation statistics.

        Returns
        -------
        evaluation_result : `dict` [`str`, `dict`, or `pandas.DataFrame`]

            - ``"constraint_violation"`` : `dict` [`str`, `float`]
                The normalized constraint violations.
                Keys are "actual", "forecast", and "adjusted".
                The value is root mean squared constraint violation divided
                by root mean squared actual value. Constraint violation of
                actuals should be close to 0.
            - ``"evaluation_df"`` : `pandas.DataFrame`, shape (m, # metrics)
                Evaluation results. DataFrame with one row for each timeseries,
                and a column for each metric listed above.

        If ``is_train``, results are stored to ``self.constraint_violation``,  ``self.evaluation_df``.
        Otherwise, they are stored to ``self.constraint_violation_test``, ``self.evaluation_df_test``.
        """
        if is_train:
            forecasts = self.forecasts
            actuals = self.actuals
            adjusted_forecasts = self.adjusted_forecasts
        else:
            if actuals_test is None:
                raise ValueError("`actuals_test` must be provided to evaluate on test set")
            self.actuals_test = actuals_test
            forecasts = self.forecasts_test
            actuals = self.actuals_test
            adjusted_forecasts = self.adjusted_forecasts_test

        forecasts = reorder_columns(forecasts, order_dict=self.order_dict)
        actuals = reorder_columns(actuals, order_dict=self.order_dict)
        forecasts_arr = np.array(forecasts)  # rows are dates (long-format), better for plotting
        actuals_arr = np.array(actuals)
        adjusted_forecasts_arr = np.array(adjusted_forecasts)

        # Constraint violation
        observation_magnitude = np.sqrt(np.mean(np.square(actuals_arr)))
        constraint_violation = {
            "adjusted": np.sqrt(np.mean(np.square(self.constraint_matrix @ adjusted_forecasts_arr.T))) / observation_magnitude,
            "forecast": np.sqrt(np.mean(np.square(self.constraint_matrix @ forecasts_arr.T))) / observation_magnitude,
            "actual": np.sqrt(np.mean(np.square(self.constraint_matrix @ actuals_arr.T))) / observation_magnitude,
        }

        # Adjusted forecast and original forecast error
        evaluation_df = pd.DataFrame({
            "Adjusted MAPE": 100 * np.nanmean(np.abs(adjusted_forecasts_arr - actuals_arr) / np.abs(actuals_arr), axis=0),
            "Base MAPE": 100 * np.nanmean(np.abs(forecasts_arr - actuals_arr) / np.abs(actuals_arr), axis=0),
            "Adjusted MedAPE": 100 * np.nanmedian(np.abs(adjusted_forecasts_arr - actuals_arr) / np.abs(actuals_arr), axis=0),
            "Base MedAPE": 100 * np.nanmedian(np.abs(forecasts_arr - actuals_arr) / np.abs(actuals_arr), axis=0),
            "Adjusted RMSE": np.sqrt(np.mean(np.square(actuals_arr - adjusted_forecasts_arr), axis=0)),
            "Base RMSE": np.sqrt(np.mean(np.square(actuals_arr - forecasts_arr), axis=0)),
        }, index=actuals.columns)

        evaluation_df["RMSE % change"] = 100 * (evaluation_df["Adjusted RMSE"] / evaluation_df["Base RMSE"] - 1.0)
        evaluation_df["MAPE pp change"] = evaluation_df["Adjusted MAPE"] - evaluation_df["Base MAPE"]
        evaluation_df["MedAPE pp change"] = evaluation_df["Adjusted MedAPE"] - evaluation_df["Base MedAPE"]
        column_order = [
            "RMSE % change",
            "MAPE pp change",
            "MedAPE pp change",
            "Base MAPE",
            "Base MedAPE",
            "Base RMSE",
            "Adjusted MAPE",
            "Adjusted MedAPE",
            "Adjusted RMSE",
        ]
        evaluation_df = evaluation_df[column_order]

        if ipython_display:
            if "IPython.core" in sys.modules:
                ICD.display(evaluation_df.round(1))
            else:
                print(evaluation_df.round(1))

        if plot:
            # Basic plots for convenience.
            num_cols = forecasts.shape[1]
            height = 6
            # See the adjustments on original scale
            plt.figure(figsize=(height * num_cols, height))
            for i, dim in enumerate(forecasts.columns):
                plt.subplot(int(f"1{num_cols}{i+1}"))
                y1 = np.array(forecasts[dim])
                y2 = np.array(adjusted_forecasts[dim])
                plt.plot(y1, label="Base forecast")
                plt.plot(y2, label="Adjusted forecast")
                plt.xlabel("time index")
                plt.ylabel("value")
                plt.title(dim)
            plt.suptitle("Base vs Adjusted Forecast")
            plt.legend()
            plt.show()

            # Relative difference
            plt.figure(figsize=(height * num_cols, height))
            for i, dim in enumerate(forecasts.columns):
                plt.subplot(int(f"1{num_cols}{i+1}"))
                y1 = np.array(forecasts[dim])
                y2 = np.array(adjusted_forecasts[dim])
                plt.plot(100 * (y2/y1 - 1))
                plt.xlabel("time index")
                plt.ylabel("% difference")
                plt.title(dim)
            plt.suptitle("% difference: Base vs Adjusted Forecast")
            plt.show()

            # Change in % error
            plt.figure(figsize=(height * num_cols, height))
            for i, dim in enumerate(forecasts.columns):
                plt.subplot(int(f"1{num_cols}{i+1}"))
                y1 = np.array(actuals[dim])
                y2 = np.array(forecasts[dim])
                y3 = np.array(adjusted_forecasts[dim])
                plt.plot(y2/y1 - 1, label='Base Forecast')
                plt.plot(y3/y1 - 1, label='Adjusted Forecast')
                plt.xlabel("time index")
                plt.ylabel("% error")
                plt.title(dim)
            plt.suptitle('Forecast Percent Error (vs actual)')
            plt.legend()
            plt.show()

        if is_train:
            self.constraint_violation = constraint_violation
            self.evaluation_df = evaluation_df
        else:
            self.constraint_violation_test = constraint_violation
            self.evaluation_df_test = evaluation_df

        return {
            "constraint_violation": constraint_violation,
            "evaluation_df": evaluation_df
        }

    def fit_transform(self, forecasts, actuals, **fit_kwargs):
        """Fits and transforms training data.

        Parameters
        ----------
        forecasts : `pandas.DataFrame`
            Forecasts to fit the adjustment. See ``fit``.
        actuals : `pandas.DataFrame`
            Actuals to fit the adjustment. See ``fit``.
        fit_kwargs : `dict`, optional
            Additional parameters to pass to ``fit``.

        Returns
        -------
        adjusted_forecasts : `pandas.DataFrame`
            Adjusted forecasts.
        """
        self.fit(forecasts, actuals, **fit_kwargs)
        self.transform()
        return self.adjusted_forecasts

    def fit_transform_evaluate(self, forecasts, actuals, fit_kwargs=None, evaluate_kwargs=None):
        """Fits, transforms, and evaluates on training data.

        Parameters
        ----------
        forecasts : `pandas.DataFrame`
            Forecasts to fit the adjustment. See ``fit``.
        actuals : `pandas.DataFrame`
            Actuals to fit the adjustment. See ``fit``.
        fit_kwargs : `dict`, optional, default None
            Additional parameters to pass to ``fit``.
        evaluate_kwargs : `dict`, optional, default None
            Additional parameters to pass to ``evaluate``.

        Returns
        -------
        evaluation_df : `pandas.DataFrame`
            Evaluation results on provided ``forecasts``.
        """
        if fit_kwargs is None:
            fit_kwargs = {}
        if evaluate_kwargs is None:
            evaluate_kwargs = {}
        self.fit(forecasts, actuals, **fit_kwargs)
        self.transform()
        self.evaluate(is_train=True, **evaluate_kwargs)
        return self.evaluation_df

    def transform_evaluate(self, forecasts_test, actuals_test, **evaluate_kwargs):
        """Transforms and evaluates on test data.

        Must call ``fit`` before calling this method.

        forecasts_test : `pandas.DataFrame`
            Forecasts to make consistent. Should be different from the training data.
        actuals_test : `pandas.DataFrame`
            Actuals to check quality of the adjustment.
        evaluate_kwargs : `dict`, optional, default None
            Additional parameters to pass to ``evaluate``.

        Returns
        -------
        evaluation_df : `pandas.DataFrame`
            Evaluation results on provided ``forecasts_test``.
        """
        self.transform(forecasts_test)
        self.evaluate(is_train=False, actuals_test=actuals_test, **evaluate_kwargs)
        return self.evaluation_df



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Estimates the empirical distribution of a variable."""

import numpy as np


def estimate_empirical_distribution(
        df,
        value_col,
        quantile_grid_size=0.05,
        quantiles=None,
        conditional_cols=None):
    """Estimates an empirical distribution for a given variable in the dataframe (df) "value_col" column
    The distribution is an approximated quantile function defined on
    the quantiles specified in "quantiles" which are values in [0, 1]
    The function returns, mean, and standard deviation (std) as well.

    :param df: the dataframe which includes the dataframe
    :param value_col: the column name for the value column of interest
    :param quantile_grid_size: the grid size for the quantiles.
        E.g. if grid_size = 0.25 then the quantiles for 0.25, 0.5, 0.75 are calculated
        If grid_size = 0.05 then then the quantiles for 0.05, 0.1, ..., 0.90, 0.95 are calculated
    :param quantiles: the probability grid for which quantiles are calculated.
        If None quantile_grid_size is used to generate a regularly spaced one.
    :param conditional_cols: the grouping variables used to condition on

    :return: The output consists of a dictionary with two dataframes
        (1) first dataframe ("ecdf_df") with columns being
            conditional_cols
            mean, sd
            quantiles (as many as specified in quantiles)
            min, max
            sample_size (count) for each combination
        (2) second dataframe ("ecdf_df_overall") has only one row and returns
            the overall mean, sd, quantiles (no conditional_cols), min, max, overall sample size (count)
    """
    if quantiles is None:
        num = int(1.0 / quantile_grid_size - 1.0)
        quantiles = np.linspace(
            quantile_grid_size,
            1 - quantile_grid_size,
            num=num)

    def quantile_summary(x):
        return tuple(np.quantile(a=x, q=quantiles))

    # Aggregates w.r.t. given columns in conditional_cols
    agg_dict = {value_col: [quantile_summary, "min", "mean", "max", "std", "count"]}
    # Creates an overall distribution by simply aggregating every value column globally.
    # This is useful for cases where no matching data is available for a given combination of values in conditional_cols
    ecdf_df_overall = df.groupby([True]*len(df), as_index=False).agg(agg_dict)
    # Flattens multi-index (result of aggregation) to have flat and descriptive column names.
    ecdf_df_overall.columns = [
        f"{value_col}_{x}" for x in ["quantile_summary", "min", "mean", "max", "std", "count"]]
    if conditional_cols is None:
        ecdf_df = ecdf_df_overall
    else:
        ecdf_df = df.groupby(conditional_cols, as_index=False).agg(agg_dict)
        # Flattens multi-index (result of aggregation) to have flat and descriptive column names.
        ecdf_df.columns = [f"{a}_{b}" if b else a for (a, b) in ecdf_df.columns]

    model = {
        "ecdf_df": ecdf_df,
        "ecdf_df_overall": ecdf_df_overall}

    return model



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen, Kaixu Yang
"""Logging functions."""

import logging
from enum import Enum

import numpy as np
import six

from greykite.common.constants import LOGGER_NAME


# Here we name the logger "LOGGER_NAME".
# We don't want to change the default behaviour of the root logger,
# which will influence the behaviour of other modules.
logger = logging.getLogger(LOGGER_NAME)


class LoggingLevelEnum(Enum):
    """Valid types of logging levels available to use."""
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0


logger.setLevel(LoggingLevelEnum.WARNING.value)  # only logs events with level at least WARNING


def pprint(params, offset=0, printer=repr):
    """Pretty print the dictionary 'params'

    Copied from sklearn.base._pprint to avoid accessing protected member of module

    Parameters
    ----------
    params : dict
        The dictionary to pretty print

    offset : int
        The offset in characters to add at the begin of each line.

    printer : callable
        The function to convert entries to strings, typically
        the builtin str or repr

    """
    # Do a multi-line justified repr:
    options = np.get_printoptions()
    np.set_printoptions(precision=5, threshold=64, edgeitems=2)
    params_list = list()
    this_line_length = offset
    line_sep = ',\n' + (1 + offset // 2) * ' '
    for i, (k, v) in enumerate(sorted(six.iteritems(params))):
        if type(v) is float:
            # use str for representing floating point numbers
            # this way we get consistent representation across
            # architectures and versions.
            this_repr = '%s=%s' % (k, str(v))
        else:
            # use repr of the rest
            this_repr = '%s=%s' % (k, printer(v))
        if len(this_repr) > 500:
            this_repr = this_repr[:300] + '...' + this_repr[-100:]
        if i > 0:
            if this_line_length + len(this_repr) >= 75 or '\n' in this_repr:
                params_list.append(line_sep)
                this_line_length = len(line_sep)
            else:
                params_list.append(', ')
                this_line_length += 2
        params_list.append(this_repr)
        this_line_length += len(this_repr)

    np.set_printoptions(**options)
    lines = ''.join(params_list)
    # Strip trailing space to avoid nightmare in doctests
    lines = '\n'.join(l.rstrip(' ') for l in lines.split('\n'))
    return lines


def log_message(message, level=LoggingLevelEnum.INFO):
    """Adds a message to logger.

    Parameters
    ----------
    message : `any`
        The message to be added to logger.
    level : `Enum`
        One of the levels in the `~greykite.common.enums.LoggingLevelEnum`.
    """
    if level.name not in list(LoggingLevelEnum.__members__):
        raise ValueError(f"{level} not found, it must be a member of the LoggingLevelEnum class.")
    logger.log(level.value, message)



import warnings
from functools import partial
from typing import Optional

import numpy as np
import pandas as pd
import plotly
from sklearn.model_selection import RandomizedSearchCV
from sklearn.pipeline import Pipeline

from greykite.common.constants import ACTUAL_COL
from greykite.common.constants import PREDICTED_COL
from greykite.common.constants import PREDICTION_BAND_COVERAGE
from greykite.common.constants import TIME_COL
from greykite.common.constants import VALUE_COL
from greykite.common.evaluation import EvaluationMetricEnum
from greykite.common.evaluation import fraction_outside_tolerance
from greykite.common.python_utils import assert_equal
from greykite.common.testing_utils import assert_eval_function_equal
from greykite.framework.constants import FRACTION_OUTSIDE_TOLERANCE_NAME
from greykite.framework.output.univariate_forecast import UnivariateForecast
from greykite.framework.pipeline.pipeline import validate_pipeline_input
from greykite.framework.pipeline.utils import get_score_func_with_aggregation
from greykite.framework.utils.result_summary import summarize_grid_search_results
from greykite.sklearn.estimator.simple_silverkite_estimator import SimpleSilverkiteEstimator
from greykite.sklearn.sklearn_scorer import _PredictScorerDF


def assert_refit(
        refit,
        expected_metric,
        expected_greater_is_better):
    """Checks if `refit` is initialized with
        metric = expected_metric,
        and greater_is_better = expected_greater_is_better.

    Parameters
    ----------
    refit : callable
        `refit` callable to check
    expected_metric : `str`
        expected metric used to define ``refit``
    expected_greater_is_better : `bool`
        expected greater_is_better used to define ``refit``
    """
    expected_index = 1 if expected_greater_is_better else 0
    assert refit(results={f"mean_test_{expected_metric}": np.array([0.0, 1.0])}) == expected_index


def assert_scoring(
        scoring,
        expected_keys=None,
        agg_periods=None,
        agg_func=None,
        relative_error_tolerance=None):
    """Checks if `scoring` has the expected keys and score functions
    defined by the other parameters.

    Parameters
    ----------
    scoring : `dict`
        ``scoring`` dictionary to check
    expected_keys : `set` [`str`] or None
        Expected keys in `scoring` dictionary.
        If None, does not check the keys.
    agg_periods : callable or None
        What was passed to `get_scoring_and_refit`
    agg_func : `int` or None
        What was passed to `get_scoring_and_refit`
    relative_error_tolerance : `float` or None
        What was passed to `get_scoring_and_refit`
        Must provide `relative_error_tolerance` to check FRACTION_OUTSIDE_TOLERANCE_NAME.
    """
    if expected_keys is not None:
        assert scoring.keys() == expected_keys
    # a few metrics to spot check
    name_func = {
        EvaluationMetricEnum.MeanAbsolutePercentError.get_metric_name(): EvaluationMetricEnum.MeanAbsolutePercentError.get_metric_func(),
        EvaluationMetricEnum.Quantile95.get_metric_name(): EvaluationMetricEnum.Quantile95.get_metric_func(),
        FRACTION_OUTSIDE_TOLERANCE_NAME: partial(fraction_outside_tolerance, rtol=relative_error_tolerance)
    }
    for name, scorer in scoring.items():
        assert isinstance(scorer, _PredictScorerDF)
        assert scorer._sign == 1  # because greater_is_better=True
        if name in name_func:
            expected_func = get_score_func_with_aggregation(
                score_func=name_func[name],
                agg_periods=agg_periods,
                agg_func=agg_func)[0]
            assert_eval_function_equal(
                scorer._score_func,
                expected_func)


def assert_proper_grid_search(
        grid_search,
        expected_grid_size=None,
        lower_bound=None,
        upper_bound=None,
        score_func=EvaluationMetricEnum.MeanAbsolutePercentError.name,
        greater_is_better=False,
        cv_report_metrics_names=None):
    """Checks fitted hyperparameter grid search result.

    Parameters
    ----------
    grid_search : `sklearn.model_selection.RandomizedSearchCV`
        Fitted RandomizedSearchCV object
    expected_grid_size : `int` or None, default None
        Expected number of options evaluated in grid search.
        If None, does not check the expected size.
    lower_bound : `float` or None, default None
        Lower bound on CV test set error.
        If None, does not check the test error.
    upper_bound : `float` or None, default None
        Upper bound on CV test set error.
        If None, does not check the test error.
    score_func : `str` or callable, default ``EvaluationMetricEnum.MeanAbsolutePercentError.name``
        Score function used to select optimal model in CV.
        The same as passed to ``forecast_pipeline`` and grid search.
        If a callable, takes arrays ``y_true``, ``y_pred`` and returns a float.
        If a string, must be either a
        `~greykite.common.evaluation.EvaluationMetricEnum` member name
        or `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.
    greater_is_better : `bool`, default False
        Whether higher values are better.
        Must be explicitly passed for testing (not derived from ``score_func``).
    cv_report_metrics_names : `list` [`str`] or None, default None
        Additional metrics besides ``metric`` calculated during CV.
        If None, no other metrics are checked in the result.

        Unlike in ``forecast_pipeline``, these are the expected names,
        in the CV output, such as:

            - ``enum.get_metric_name()``
            - ``"CUSTOM_SCORE_FUNC_NAME"``
            - ``"FRACTION_OUTSIDE_TOLERANCE_NAME"``

    Raises
    ------
    AssertionError
        If grid search did not run as expected.
    """
    _, _, short_name = get_score_func_with_aggregation(
        score_func=score_func,  # string or callable
        greater_is_better=greater_is_better,
        # Dummy value, doesn't matter because we ignore the returned `score_func`
        relative_error_tolerance=0.01)
    # attributes are populated
    assert hasattr(grid_search, "best_estimator_")
    assert hasattr(grid_search, "cv_results_")
    if callable(grid_search.refit):
        # `grid_search.refit` is a callable if `grid_search` comes from
        # `forecast_pipeline`.
        # Checks if `best_index_` and `refit` match `metric` and `greater_is_better`.
        assert grid_search.best_index_ == grid_search.refit(grid_search.cv_results_)
        split_scores = grid_search.cv_results_[f"mean_test_{short_name}"]
        expected_best_score = max(split_scores) if greater_is_better else min(split_scores)
        assert split_scores[grid_search.best_index_] == expected_best_score
        assert split_scores[grid_search.best_index_] is not None
        assert not np.isnan(split_scores[grid_search.best_index_])
        assert_refit(
            grid_search.refit,
            expected_metric=short_name,
            expected_greater_is_better=greater_is_better)
    elif grid_search.refit is True:
        # In single metric evaluation, refit_metric is "score".
        short_name = "score"
        # `best_score_` is populated, and the optimal score is the highest
        # test set score. Metrics where `greater_is_better=False` are
        # assumed to be negated in the ``scoring`` parameter so that
        # higher values are better.
        assert hasattr(grid_search, "best_score_")
        best_score = grid_search.best_score_
        test_scores = grid_search.cv_results_[f"mean_test_{short_name}"]
        best_score2 = test_scores[grid_search.best_index_]
        assert best_score == max(test_scores)
        assert best_score2 == max(test_scores)

    if expected_grid_size is not None:
        assert len(grid_search.cv_results_[f"mean_test_{short_name}"]) == expected_grid_size
    # Parameters are populated
    assert_equal(grid_search.cv_results_["params"][grid_search.best_index_], grid_search.best_params_)

    # All metrics are computed
    if cv_report_metrics_names is None:
        cv_report_metrics_names = []
    for expected_metric in cv_report_metrics_names + [short_name]:
        assert f"mean_test_{expected_metric}" in grid_search.cv_results_.keys()
        assert f"std_test_{expected_metric}" in grid_search.cv_results_.keys()
        assert f"mean_train_{expected_metric}" in grid_search.cv_results_.keys()
        assert f"std_train_{expected_metric}" in grid_search.cv_results_.keys()

    if lower_bound is not None or upper_bound is not None:
        grid_results = summarize_grid_search_results(grid_search, score_func=score_func)
        if lower_bound is not None:
            assert all(grid_results[f"mean_test_{short_name}"] >= lower_bound)
        if upper_bound is not None:
            assert all(grid_results[f"mean_test_{short_name}"] <= upper_bound)


def check_forecast_pipeline_result(
        result,
        coverage=0.95,
        strategy=None,
        interactive=False,
        expected_grid_size=None,
        lower_bound_cv=None,
        upper_bound_cv=None,
        score_func=EvaluationMetricEnum.MeanAbsolutePercentError.name,
        greater_is_better=False,
        cv_report_metrics_names=None,
        relative_error_tolerance=None):
    """Helper function that validates forecast_pipeline output.
    Raises an AssertionError is results do not match the expected values.

    Parameters
    ----------
    result : :class:`~greykite.framework.pipeline.pipeline.ForecastResult`
        ``forecast_pipeline`` output to check
    coverage : `float` or None, default 0.95
        The ``coverage`` passed to ``forecast_pipeline``
    strategy : `str` or None, default None
        Null model strategy.
        If None, not checked.
    interactive : `bool`, default False
        Whether to plot and print results.
    expected_grid_size : `int` or None, default None
        Expected number of options evaluated in grid search.
        If None, does not check the expected size.
    lower_bound_cv : `float` or None, default None
        Lower bound on CV test set error.
        If None, does not check the test error.
    upper_bound_cv : `float` or None, default None
        Upper bound on CV test set error.
        If None, does not check the test error.
    score_func : `str` or callable, default ``EvaluationMetricEnum.MeanAbsolutePercentError.name``
        Score function used to select optimal model in CV.
        The same as passed to ``forecast_pipeline`` and grid search.
        If a callable, takes arrays ``y_true``, ``y_pred`` and returns a float.
        If a string, must be either a
        `~greykite.common.evaluation.EvaluationMetricEnum` member name
        or `~greykite.common.constants.FRACTION_OUTSIDE_TOLERANCE`.
    greater_is_better : `bool`, default False
        Whether higher values are better.
        Must be explicitly passed for testing (not derived from ``score_func``).
    cv_report_metrics_names : `list` [`str`] or None, default None
        Additional metrics besides ``metric`` calculated during CV.
        If None, no other metrics are checked in the result.

        Unlike in ``forecast_pipeline``, these are the expected names,
        in the CV output, such as:

            - ``enum.get_metric_name()``
            - ``"CUSTOM_SCORE_FUNC_NAME"``
            - ``"FRACTION_OUTSIDE_TOLERANCE_NAME"``
    relative_error_tolerance : `float` or None
        The ``relative_error_tolerance`` passed to ``forecast_pipeline``
    """
    assert isinstance(result.grid_search, RandomizedSearchCV)
    assert isinstance(result.model, Pipeline)
    assert isinstance(result.backtest, UnivariateForecast)
    assert isinstance(result.forecast, UnivariateForecast)

    assert_proper_grid_search(
        result.grid_search,
        expected_grid_size=expected_grid_size,
        lower_bound=lower_bound_cv,
        upper_bound=upper_bound_cv,
        score_func=score_func,
        greater_is_better=greater_is_better,
        cv_report_metrics_names=cv_report_metrics_names)

    ts = result.timeseries
    assert ts.df[VALUE_COL].equals(ts.y)
    assert result.backtest.train_evaluation is not None
    assert result.backtest.test_evaluation is not None
    if coverage is None:
        assert result.forecast.coverage is None
        assert result.backtest.coverage is None
        assert result.backtest.train_evaluation[PREDICTION_BAND_COVERAGE] is None
        assert result.backtest.test_evaluation[PREDICTION_BAND_COVERAGE] is None
        expected_cols = [TIME_COL, ACTUAL_COL, PREDICTED_COL]
        assert list(result.backtest.df.columns) == expected_cols
        assert list(result.forecast.df.columns) == expected_cols
    else:
        assert round(result.forecast.coverage, 3) == round(coverage, 3)
        assert round(result.backtest.coverage, 3) == round(coverage, 3)
        assert result.backtest.train_evaluation[PREDICTION_BAND_COVERAGE] is not None
        assert result.backtest.test_evaluation[PREDICTION_BAND_COVERAGE] is not None

    assert result.forecast.train_evaluation is not None

    # Tests if null model params are set for CV
    estimator = result.model.steps[-1][-1]
    if estimator.null_model is not None and strategy is not None:
        assert estimator.null_model.strategy == strategy

    # Tests if relative_error_tolerance is set for backtest/forecast
    if relative_error_tolerance is not None:
        assert result.backtest.relative_error_tolerance == relative_error_tolerance
        assert result.forecast.relative_error_tolerance == relative_error_tolerance

    if interactive:
        print("backtest_train_evaluation", result.backtest.train_evaluation)
        print("backtest_test_evaluation", result.backtest.test_evaluation)
        print("forecast_train_evaluation", result.forecast.train_evaluation)
        print("forecast_test_evaluation", result.forecast.test_evaluation)
        print(summarize_grid_search_results(
            result.grid_search,
            score_func=score_func,
            score_func_greater_is_better=greater_is_better))
        plotly.offline.plot(ts.plot())
        plotly.offline.plot(result.backtest.plot())
        plotly.offline.plot(result.forecast.plot())


def assert_basic_pipeline_equal(actual: Pipeline, expected: Pipeline):
    """Asserts that the two pipelines are equal
    The Pipelines should be created by `get_basic_pipeline`.
    """
    # checks features
    actual_params = actual.get_params()
    expected_params = expected.get_params()
    check_keys = [
        'input__date__select_date__column_names',
        'input__response__select_val__column_names',
        'input__response__outlier__use_fit_baseline',
        'input__response__outlier__z_cutoff',
        'input__response__null__impute_algorithm',
        'input__response__null__impute_all',
        'input__response__null__impute_params',
        'input__response__null__max_frac',
        'input__regressors_numeric__select_reg__column_names',
        'input__regressors_numeric__select_reg_numeric__exclude',
        'input__regressors_numeric__select_reg_numeric__include',
        'input__regressors_numeric__outlier__use_fit_baseline',
        'input__regressors_numeric__outlier__z_cutoff',
        'input__regressors_numeric__normalize__normalize_algorithm',
        'input__regressors_numeric__normalize__normalize_params',
        'input__regressors_numeric__null__impute_algorithm',
        'input__regressors_numeric__null__impute_all',
        'input__regressors_numeric__null__impute_params',
        'input__regressors_numeric__null__max_frac',
        'input__regressors_other__select_reg__column_names',
        'input__regressors_other__select_reg_non_numeric__exclude',
        'input__regressors_other__select_reg_non_numeric__include',
        'degenerate__drop_degenerate'
    ]
    for key in check_keys:
        assert actual_params[key] == expected_params[key],\
            f"{key} is different, found {actual_params[key]}, expected {expected_params[key]}"

    # checks estimator
    actual_estimator = actual.steps[-1][-1]
    expected_estimator = expected.steps[-1][-1]
    assert isinstance(actual_estimator, type(expected_estimator))
    actual_estimator_dict = actual_estimator.__dict__.copy()
    expected_estimator_dict = expected_estimator.__dict__.copy()
    del actual_estimator_dict["null_model"]
    del expected_estimator_dict["null_model"]
    actual_estimator_dict.pop("silverkite", None)
    expected_estimator_dict.pop("silverkite", None)
    actual_estimator_dict.pop("silverkite_diagnostics", None)
    expected_estimator_dict.pop("silverkite_diagnostics", None)
    actual_score_func = actual_estimator_dict.pop("score_func")
    expected_score_func = expected_estimator_dict.pop("score_func")
    assert_equal(actual_estimator_dict, expected_estimator_dict)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        assert_eval_function_equal(actual_score_func, expected_score_func)


def assert_forecast_pipeline_result_equal(actual, expected, is_silverkite=True, rel=1e-5):
    """Raises an AssertionError if the two forecast pipeline results are not equal"""
    if is_silverkite:
        # checks training matrix, predictors, evaluation splits,
        # and fit parameters
        keys_to_check = {
            "time_col",
            "value_col",
            "origin_for_time_vars",
            "daily_event_df_dict",
            "changepoint_values",
            "continuous_time_col",
            # "growth_func",  # skip, hard to check function equality
            # "fs_func",
            # "autoreg_func",
            "max_lag_order",
            "uncertainty_dict",
            "pred_cols",
            "last_date_for_fit",
            "x_mat"  # training matrix
        }
        simple_model_dict = actual.model.steps[-1][-1].model_dict
        model_dict = expected.model.steps[-1][-1].model_dict
        assert_equal(
            {k: v for k, v in simple_model_dict.items() if k in keys_to_check},
            {k: v for k, v in model_dict.items() if k in keys_to_check},
            ignore_list_order=True
        )
        simple_fit_algorithm_dict = actual.model.steps[-1][-1].fit_algorithm_dict
        fit_algorithm_dict = expected.model.steps[-1][-1].fit_algorithm_dict
        assert simple_fit_algorithm_dict == fit_algorithm_dict

    # checks evaluation result on backtest and forecast,
    # including prediction interval and null model metrics
    if actual.backtest is None:
        assert expected.backtest is None
    else:
        assert_equal(
            actual.backtest.train_evaluation,
            expected.backtest.train_evaluation,
            rel=rel)
        assert_equal(
            actual.backtest.test_evaluation,
            expected.backtest.test_evaluation,
            rel=rel)
    assert_equal(
        actual.forecast.train_evaluation,
        expected.forecast.train_evaluation,
        rel=rel)
    assert_equal(
        actual.forecast.test_evaluation,
        expected.forecast.test_evaluation,
        rel=rel)
    assert_equal(
        actual.grid_search.best_index_,
        expected.grid_search.best_index_,
        rel=rel)
    # checks CV search parameters
    assert_equal(
        actual.grid_search.cv.__dict__,
        expected.grid_search.cv.__dict__
    )


@validate_pipeline_input
def mock_pipeline(
        # The arguments and defaults should be identical to forecast_pipeline() function
        # input
        df: pd.DataFrame,
        time_col=TIME_COL,
        value_col=VALUE_COL,
        date_format=None,
        tz=None,
        freq=None,
        train_end_date=None,
        anomaly_info=None,
        # model
        pipeline=None,
        regressor_cols=None,
        estimator=SimpleSilverkiteEstimator(),
        hyperparameter_grid=None,
        hyperparameter_budget=None,
        n_jobs=1,
        verbose=1,
        # forecast
        forecast_horizon=None,
        coverage: Optional[float] = 0.95,
        test_horizon=None,
        periods_between_train_test=None,
        agg_periods=None,
        agg_func=None,
        # evaluation
        score_func=EvaluationMetricEnum.MeanSquaredError.name,
        score_func_greater_is_better=False,
        cv_report_metrics=None,
        null_model_params=None,
        relative_error_tolerance: Optional[float] = 0.05,
        # CV
        cv_horizon=None,
        cv_min_train_periods=None,
        cv_expanding_window=False,
        cv_use_most_recent_splits=False,
        cv_periods_between_splits=None,
        cv_periods_between_train_test=0,
        cv_max_splits=3):
    """Create and returns custom pipeline parameters"""
    return {
        "df": df,
        "time_col": time_col,
        "value_col": value_col,
        "date_format": date_format,
        "tz": tz,
        "freq": freq,
        "train_end_date": train_end_date,
        "anomaly_info": anomaly_info,
        "pipeline": pipeline,
        "regressor_cols": regressor_cols,
        "estimator": estimator,
        "hyperparameter_grid": hyperparameter_grid,
        "hyperparameter_budget": hyperparameter_budget,
        "n_jobs": n_jobs,
        "verbose": verbose,
        "forecast_horizon": forecast_horizon,
        "coverage": coverage,
        "test_horizon": test_horizon,
        "periods_between_train_test": periods_between_train_test,
        "agg_periods": agg_periods,
        "agg_func": agg_func,
        "score_func": score_func,
        "score_func_greater_is_better": score_func_greater_is_better,
        "cv_report_metrics": cv_report_metrics,
        "null_model_params": null_model_params,
        "relative_error_tolerance": relative_error_tolerance,
        "cv_horizon": cv_horizon,
        "cv_min_train_periods": cv_min_train_periods,
        "cv_expanding_window": cv_expanding_window,
        "cv_use_most_recent_splits": cv_use_most_recent_splits,
        "cv_periods_between_splits": cv_periods_between_splits,
        "cv_periods_between_train_test": cv_periods_between_train_test,
        "cv_max_splits": cv_max_splits,
    }



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Defines an image scraper for sphinx-gallery
https://sphinx-gallery.github.io/
which can be used to show plotly output in documentation.
Adapted from the functions in plotly v4.6.0 to support:

    1. List of examples_dirs, gallery_dirs
    2. Do not render png thumbnail to avoid dependencies

Compatible with plotly 3.10.0.
"""

import inspect
import os
import shutil
from glob import glob
from pathlib import Path

import plotly
from plotly.io import write_html
from plotly.io import write_image
from plotly.io._base_renderers import ExternalRenderer
from plotly.tools import return_figure_from_figure_or_data


class SphinxGalleryRenderer(ExternalRenderer):
    """Original class: `from plotly.io._base_renderers import SphinxGalleryRenderer`
    Modified to add `render_png` parameter.
    """
    def render(self, fig_dict):
        """Called by plotly.io.show"""
        stack = inspect.stack()
        # Name of .py example script from which plot function was called
        try:
            filename = stack[3].filename  # let's hope this is robust...
        except (IndexError, AttributeError):  # python 2
            filename = stack[3][1]
        # NB: this works for multiple plots in the same file
        # It appears they are executed sequentially
        filename_root, _ = os.path.splitext(filename)
        filename_html = filename_root + ".html"
        filename_png = filename_root + ".png"
        _ = write_html(fig_dict, file=filename_html)

        # Whether to render the html image as a png for the thumbnail
        render_png = False
        if render_png:
            # Requires plotly-orca and xfvb
            # https://github.com/plotly/orca
            plotly.io.orca.config.use_xvfb = True
            figure = return_figure_from_figure_or_data(fig_dict, True)
            write_image(figure, filename_png)
        else:
            # The thumbnail isn't important, so we use a default image
            # Assumes the default image is one level above the .py file
            filename_default_image = Path(filename_root).parents[1].joinpath('default_thumb.png')
            shutil.copyfile(filename_default_image, filename_png)


# Original code: https://github.com/plotly/plotly.py/issues/1459
# Register for use by name
sphinx_renderer = SphinxGalleryRenderer()
plotly.io.renderers['sphinx_gallery'] = sphinx_renderer

# Set as default
plotly.io.renderers.default = 'sphinx_gallery'


def plotly_sg_scraper(block, block_vars, gallery_conf, **kwargs):
    """Scrape Plotly figures for galleries of examples using
    sphinx-gallery.
    Examples should use ``plotly.io.show()`` to display the figure with
    the custom sphinx_gallery renderer.
    Since the sphinx_gallery renderer generates both html and static png
    files, we simply crawl these files and give them the appropriate path.

    Original function:
        from plotly.io._sg_scraper import plotly_sg_scraper
    Which is based on:
        https://sphinx-gallery.github.io/stable/advanced.html#example-2-detecting-image-files-on-disk
    Modified to handle the case where examples_dirs is a list of more than one item.
        Original code failed to render plots for multiple directories:
            - Only the rendered the first directory
            - If multiple directories were specified, matplotlib plots in first directory were not rendered

    Parameters
    ----------
    block : tuple
        A tuple containing the (label, content, line_number) of the block.
    block_vars : dict
        Dict of block variables.
    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery
    **kwargs : dict
        Additional keyword arguments to pass to
        :meth:`~matplotlib.figure.Figure.savefig`, e.g. ``format='svg'``.
        The ``format`` kwarg in particular is used to set the file extension
        of the output file (currently only 'png' and 'svg' are supported).
    Returns
    -------
    rst : str
        The ReSTructuredText that will be rendered to HTML containing
        the images.
    Notes
    -----
    Add this function to the image scrapers
    """
    # e.g. '/home/user/Documents/greykite/module/docs/gallery/subgallery/images/sphx_glr_plot_mymodule_{0:03}.png'
    # where 'gallery/subgallery' is an element specified in gallery_conf["gallery_dirs"].
    image_path_iterator = block_vars["image_path_iterator"]

    # Finds the examples_dirs directory with the files
    # to move to the locations specified by image_path_iterator
    examples_dirs = gallery_conf["examples_dirs"]
    if isinstance(examples_dirs, (list, tuple)):
        # Gets `gallery_dirs` item corresponding to the ``image_path_iterator``
        # example: '/home/user/Documents/mp/module/docs/gallery/subgallery'
        abs_gallery_path = Path(image_path_iterator.image_path).parents[1]
        # example: 'gallery/subgallery'
        rel_gallery_path = os.path.relpath(abs_gallery_path, gallery_conf["src_dir"])
        # Maps relative paths in gallery_dirs to those in examples_dirs
        gallery_to_examples = dict(zip(gallery_conf["gallery_dirs"], examples_dirs))
        # Gets `example_dirs` item corresponding to the gallery_dirs item
        rel_example_path = gallery_to_examples[rel_gallery_path]
        # Converts to absolute example path
        # `rel_example_path` is the path relative to gallery_conf["src_dir"]
        # (/home/user/Documents/mp/module/docs) but the current working directory
        # could be /home/user/Documents/mp or something else
        abs_example_path = os.path.join(gallery_conf["src_dir"], rel_example_path)

    # Copies files from rel_example_path to the proper location in gallery_dirs
    pngs = sorted(glob(os.path.join(abs_example_path, "*.png")))
    htmls = sorted(glob(os.path.join(abs_example_path, "*.html")))
    image_names = list()
    seen = set()

    for html, png in zip(htmls, pngs):
        if png not in seen:
            seen |= set(png)
            # the incrementor simply increments the filename as long as html/png are found
            # _001.png, _002.png, _003.png, etc.
            this_image_path_png = next(image_path_iterator)
            this_image_path_html = os.path.splitext(this_image_path_png)[0] + ".html"
            image_names.append(this_image_path_html)
            shutil.move(png, this_image_path_png)
            shutil.move(html, this_image_path_html)
    # Use the `figure_rst` helper function to generate rST for image files
    return figure_rst(image_names, gallery_conf["src_dir"])


def figure_rst(figure_list, sources_dir):
    """Generate RST for a list of image filenames.
    Depending on whether we have one or more figures, we use a
    single rst call to 'image' or a horizontal list.

    Original functions:
        from sphinx_gallery.gen_rst import figure_rst
        from plotly.io._sg_scraper import figure_rst
    Modified rST directives from image to raw html

    Parameters
    ----------
    figure_list : list
        List of strings of the figures' absolute paths.
    sources_dir : str
        absolute path of Sphinx documentation sources
    Returns
    -------
    images_rst : str
        rst code to embed the images in the document
    """
    # figure_paths = [os.path.relpath(figure_path, sources_dir)
    #                 .replace(os.sep, '/').lstrip('/')
    #                 for figure_path in figure_list]
    figure_paths = figure_list
    images_rst = ""
    if len(figure_paths) == 1:
        figure_name = figure_paths[0]
        images_rst = SINGLE_HTML % figure_name
    elif len(figure_paths) > 1:
        images_rst = HLIST_HEADER
        for figure_name in figure_paths:
            images_rst += HLIST_HTML_TEMPLATE % figure_name

    return images_rst


# The following strings are used when we have several pictures: we use
# an html div tag that our CSS uses to turn the lists into horizontal
# lists.
HLIST_HEADER = """"""

HLIST_HTML_TEMPLATE = """
    *
        .. raw:: html
            :file: %s
"""

SINGLE_HTML = """
.. raw:: html
    :file: %s
"""



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""This file applies and runs pre-packaged model templates.

A model template provides a wrapper around
:func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`
that provides suitable parameters for modeling different types of input data
(frequency, forecast horizon, etc.).

Each model template provides default parameters to call
:func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`,
most importantly:

    * ``estimator_name``,
    * ``hyperparameter_grid``,
    * ``pipeline``

These defaults can be customized by the other parameters of a
:class:`~greykite.framework.templates.model_templates.ForecastConfig`.

Thus, a model template defines a model, whereas
:func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`
takes a model through the training and validation steps.

Thus, model templates make it easy to create and tune a forecast.
Users are advised to start with the provided templates and tune the
model component parameters as necessary.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Type

from greykite.framework.templates.prophet_template import ProphetTemplate
from greykite.framework.templates.silverkite_template import SilverkiteTemplate
from greykite.framework.templates.simple_silverkite_template import SimpleSilverkiteTemplate
from greykite.framework.templates.template_interface import TemplateInterface


@dataclass
class ModelTemplate:
    """A model template consists of a template class,
    a description, and a name.

    This class holds the template class and description.
    The model template name is the member name in
    `greykite.framework.templates.model_templates.ModelTemplateEnum`.
    """
    template_class: Type[TemplateInterface]
    """A class that implements the template interface."""
    description: str
    """A description of the model template."""


class ModelTemplateEnum(Enum):
    """Available model templates.

    Enumerates the possible values for the ``model_template`` attribute of
    :class:`~greykite.framework.templates.model_templates.ForecastConfig`.

    The value has type `~greykite.framework.templates.model_templates.ModelTemplate` which contains:

        - the template class that recognizes the model_template. Template classes implement the
          `~greykite.framework.templates.template_interface.TemplateInterface` interface.
        - a plain-text description of what the model_template is for,

    The description should be unique across enum members. The template class
    can be shared, because a template class can recognize multiple model templates.
    For example, the same template class may use different default values for
    ``ForecastConfig.model_components_param`` depending on ``ForecastConfig.model_template``.

    Notes
    -----
    The template classes
    `~greykite.framework.templates.silverkite_template.SilverkiteTemplate`
    and `~greykite.framework.templates.prophet_template.ProphetTemplate`
    recognize only the model templates explicitly enumerated here.

    However, the `~greykite.framework.templates.simple_silverkite_template.SimpleSilverkiteTemplate`
    template class allows additional model templates to be specified generically.
    Any object of type `~greykite.framework.templates.simple_silverkite_template_config.SimpleSilverkiteTemplateOptions`
    can be used as the model_template.
    These generic model templates are valid but not enumerated here.
    """
    SILVERKITE = ModelTemplate(
        template_class=SimpleSilverkiteTemplate,
        description="Silverkite model with automatic growth, seasonality, holidays, "
                    "and interactions. Best for hourly and daily frequencies."
                    "Uses `SimpleSilverkiteEstimator`.")
    """Silverkite model with automatic growth, seasonality, holidays,
    and interactions. Best for hourly and daily frequencies.
    Uses `SimpleSilverkiteEstimator`.
    """
    SILVERKITE_DAILY_90 = ModelTemplate(
        template_class=SimpleSilverkiteTemplate,
        description="Silverkite model specifically tuned for daily data with 90 days forecast horizon. "
                    "Contains 4 hyperparameter combinations for grid search. "
                    "Uses `SimpleSilverkiteEstimator`.")
    """Silverkite model specifically tuned for daily data with 90 days forecast horizon.
    Contains 4 hyperparameter combinations for grid search.
    Uses `SimpleSilverkiteEstimator`.
    """
    SILVERKITE_WEEKLY = ModelTemplate(
        template_class=SimpleSilverkiteTemplate,
        description="Silverkite model specifically tuned for weekly data. "
                    "Contains 4 hyperparameter combinations for grid search. "
                    "Uses `SimpleSilverkiteEstimator`.")
    """Silverkite model specifically tuned for weekly data.
    Contains 4 hyperparameter combinations for grid search.
    Uses `SimpleSilverkiteEstimator`.
    """
    SILVERKITE_HOURLY_1 = ModelTemplate(
        template_class=SimpleSilverkiteTemplate,
        description="Silverkite model specifically tuned for hourly data with 1 hour forecast horizon. "
                    "Contains 4 hyperparameter combinations for grid search. "
                    "Uses `SimpleSilverkiteEstimator`.")
    """Silverkite model specifically tuned for hourly data with 1 hour forecast horizon.
    Contains 4 hyperparameter combinations for grid search.
    Uses `SimpleSilverkiteEstimator`."""
    SILVERKITE_HOURLY_24 = ModelTemplate(
        template_class=SimpleSilverkiteTemplate,
        description="Silverkite model specifically tuned for hourly data with 24 hours (1 day) forecast horizon. "
                    "Contains 4 hyperparameter combinations for grid search. "
                    "Uses `SimpleSilverkiteEstimator`.")
    """Silverkite model specifically tuned for hourly data with 24 hours (1 day) forecast horizon.
    Contains 4 hyperparameter combinations for grid search.
    Uses `SimpleSilverkiteEstimator`."""
    SILVERKITE_HOURLY_168 = ModelTemplate(
        template_class=SimpleSilverkiteTemplate,
        description="Silverkite model specifically tuned for hourly data with 168 hours (1 week) forecast horizon. "
                    "Contains 4 hyperparameter combinations for grid search. "
                    "Uses `SimpleSilverkiteEstimator`.")
    """Silverkite model specifically tuned for hourly data with 168 hours (1 week) forecast horizon.
    Contains 4 hyperparameter combinations for grid search.
    Uses `SimpleSilverkiteEstimator`."""
    SILVERKITE_HOURLY_336 = ModelTemplate(
        template_class=SimpleSilverkiteTemplate,
        description="Silverkite model specifically tuned for hourly data with 336 hours (2 weeks) forecast horizon. "
                    "Contains 4 hyperparameter combinations for grid search. "
                    "Uses `SimpleSilverkiteEstimator`.")
    """Silverkite model specifically tuned for hourly data with 336 hours (2 weeks) forecast horizon.
    Contains 4 hyperparameter combinations for grid search.
    Uses `SimpleSilverkiteEstimator`.
    """
    SILVERKITE_EMPTY = ModelTemplate(
        template_class=SimpleSilverkiteTemplate,
        description="Silverkite model with no component included by default. Fits only a constant intercept. "
                    "Select and customize this template to add only the terms you want. "
                    "Uses `SimpleSilverkiteEstimator`.")
    """Silverkite model with no component included by default. Fits only a constant intercept.
    Select and customize this template to add only the terms you want.
    Uses `SimpleSilverkiteEstimator`.
    """
    SK = ModelTemplate(
        template_class=SilverkiteTemplate,
        description="Silverkite model with low-level interface. For flexible model tuning "
                    "if SILVERKITE template is not flexible enough. Not for use out-of-the-box: "
                    "customization is needed for good performance. Uses `SilverkiteEstimator`.")
    """Silverkite model with low-level interface. For flexible model tuning
    if SILVERKITE template is not flexible enough. Not for use out-of-the-box:
    customization is needed for good performance. Uses `SilverkiteEstimator`.
    """
    PROPHET = ModelTemplate(
        template_class=ProphetTemplate,
        description="Prophet model with growth, seasonality, holidays, additional regressors "
                    "and prediction intervals. Uses `ProphetEstimator`.")
    """Prophet model with growth, seasonality, holidays, additional regressors
    and prediction intervals. Uses `ProphetEstimator`."""



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Functions to impute missing data using past data."""
import numpy as np

from greykite.common.features.timeseries_lags import build_agg_lag_df


def impute_with_lags(
        df,
        value_col,
        orders,
        agg_func=np.mean,
        iter_num=1):
    """A function to impute timeseries values (given in ``df``) and in ``value_col``
    with chosen lagged values or an aggregated of those.
    For example for daily data one could use the 7th lag to impute using the
    value of the same day of past week as opposed to the closest value available
    which can be inferior for business related timeseries.

    The imputation can be done multiple times by specifying ``iter_num``
    to decrease the number of missing in some cases.
    Note that there are no guarantees to impute all missing values with this
    method by design.
    However the original number of missing values and the final number of missing
    values are returned by the function along with the imputed dataframe.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Input dataframe which must include `value_col` as a column.
    value_col : `str`
        The column name in ``df`` representing the values of the timeseries.
    orders : list of `int`
        The lag orders to be used for aggregation.
    agg_func : callable, default `np.mean`
        `pandas.Series` -> `float`
        An aggregation function to aggregate the chosen lags.
    iter_num : `int`, default `1`
        Maximum number of iterations to impute the series.
        Each iteration represent an imputation of the series using the provided
        lag orders (``orders``) and return an imputed dataframe.
        It might be the case that with one iterations some values
        are not imputed but with more iterations one can achieve more
        imputed values.

    Returns
    -------
    impute_info : `dict`
        A dictionary with following items:

        "df" : `pandas.DataFrame`
            A dataframe with the imputed values.
        "initial_missing_num" : `int`
            Initial number of missing values.
        "final_missing_num" : `int`
            Final number of missing values after imputations.
    """
    df = df.copy()
    nan_index = df[value_col].isnull()
    missing_num = sum(nan_index)
    initial_missing_num = missing_num

    def nan_agg_func(x):
        """To avoid warnings due to aggregation performed on sub-series
        with only NAs, we define an internal modified version of ``agg_func``
        """
        x = x.dropna()
        if len(x) == 0:
            return np.nan
        else:
            return agg_func(x)

    i = 0
    while i < iter_num and missing_num > 0:
        agg_lag_info = build_agg_lag_df(
            value_col=value_col,
            df=df,
            orders_list=[orders],
            interval_list=[],
            agg_func=nan_agg_func,
            agg_name="avglag")

        agg_lag_df = agg_lag_info["agg_lag_df"]
        # In this case the lagged dataframe will have only one column
        # which can be extracted from the list given in the ``"col_names"`` item
        agg_col_name = agg_lag_info["col_names"][0]
        df.loc[nan_index, value_col] = agg_lag_df.loc[nan_index, agg_col_name]
        nan_index = df[value_col].isnull()
        missing_num = sum(nan_index)
        i += 1

    return {
        "df": df,
        "initial_missing_num": initial_missing_num,
        "final_missing_num": missing_num}


def impute_with_lags_multi(
        df,
        orders,
        agg_func=np.mean,
        iter_num=1,
        cols=None):
    """Imputes every column of ``df`` using
    `~greykite.common.features.timeseries_impute.impute_with_lags`.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Input dataframe which must include `value_col` as a column.
    orders : list of `int`
        The lag orders to be used for aggregation.
    agg_func : callable, default `np.mean`
        `pandas.Series` -> `float`
        An aggregation function to aggregate the chosen lags.
    iter_num : `int`, default `1`
        Maximum number of iterations to impute the series.
        Each iteration represent an imputation of the series using the provided
        lag orders (``orders``) and return an imputed dataframe.
        It might be the case that with one iterations some values
        are not imputed but with more iterations one can achieve more
        imputed values.
    cols : `list` [`str`] or None, default None
        Which columns to impute. If None, imputes all columns.

    Returns
    -------
    impute_info : `dict`
        A dictionary with following items:

        "df" : `pandas.DataFrame`
            A dataframe with the imputed values.
        "missing_info" : `dict`
            Dictionary with information about the missing info.

            Key = name of a column in ``df``
            Value = dictionary containing:

                "initial_missing_num" : `int`
                    Initial number of missing values.
                "final_missing_num" : `int`
                    Final number of missing values after imputation.
    """
    if cols is None:
        cols = df.columns
    impute_info = {
        "df": df.copy(),
        "missing_info": {}
    }
    for col in cols:
        # updates each column with the imputed values
        col_impute_info = impute_with_lags(
            df=df,
            value_col=col,
            orders=orders,
            agg_func=agg_func,
            iter_num=iter_num,
        )
        impute_info["df"][col] = col_impute_info["df"][col]
        impute_info["missing_info"][col] = {
            "initial_missing_num": col_impute_info["initial_missing_num"],
            "final_missing_num": col_impute_info["final_missing_num"],
        }
    return impute_info
